from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    DefaultDict,
    Generator,
    Iterable,
    Iterator,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

import numpy as np
import pandas as pd
from tqdm import tqdm

from eir.data_load.data_source_modules.common_utils import add_id_to_samples
from eir.data_load.label_setup import get_file_path_iterator
from eir.setup.input_setup_modules.setup_sequence import get_sequence_split_function
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.data_load.data_utils import Sample

logger = get_logger(name=__name__)


def get_file_sample_id_iterator(
    data_source: str, ids_to_keep: Union[None, Sequence[str]]
) -> Generator[Tuple[Any, str], None, None]:
    def _id_from_filename(file: Path) -> str:
        return file.stem

    def _filter_ids_callable(item: Path, sample_id: str) -> bool:
        assert ids_to_keep is not None
        if sample_id in ids_to_keep:
            return True
        return False

    base_file_iterator = get_file_path_iterator(
        data_source=Path(data_source), validate=False
    )

    sample_id_and_file_iterator = _get_sample_id_data_iterator(
        base_iterator=base_file_iterator, id_callable=_id_from_filename
    )

    if ids_to_keep:
        final_iterator = _get_filter_iterator(
            base_iterator=sample_id_and_file_iterator,
            filter_callable=_filter_ids_callable,
        )
    else:
        final_iterator = sample_id_and_file_iterator

    yield from final_iterator


def _get_sample_id_data_iterator(
    base_iterator: Iterable[Path], id_callable: Callable[[Path], str]
) -> Generator[Tuple[Path, str], None, None]:
    for item in base_iterator:
        sample_id = id_callable(item)
        yield item, sample_id


def _get_filter_iterator(
    base_iterator: Iterable[Tuple[Path, str]],
    filter_callable: Callable[[Path, str], bool],
) -> Generator[Any, None, None]:
    for item in base_iterator:
        if filter_callable(*item):
            yield item


def get_file_sample_id_iterator_basic(
    data_source: str,
    ids_to_keep: Union[None, Set[str]],
) -> Generator[Tuple[str, Path], None, None]:
    base_file_iterator = get_file_path_iterator(
        data_source=Path(data_source), validate=False
    )

    for file in base_file_iterator:
        sample_id = file.stem

        if ids_to_keep:
            if sample_id in ids_to_keep:
                yield sample_id, file
        else:
            yield sample_id, file


def add_sequence_data_from_csv_to_samples(
    input_source: str,
    samples: DefaultDict[str, "Sample"],
    ids_to_keep: Union[None, Set[str]],
    split_on: Optional[str],
    encode_func: Callable,
    input_name: str = "CSV File Data",
) -> DefaultDict[str, "Sample"]:
    logger.info(
        "Loading sequence data from CSV file %s. Note that this will "
        "load all the sequence data into memory.",
        input_source,
    )

    split_func = get_sequence_split_function(split_on=split_on)
    csv_sequence_iterator = get_csv_id_sequence_iterator(
        data_source=input_source, ids_to_keep=ids_to_keep
    )
    file_iterator_tqdm = tqdm(csv_sequence_iterator, desc=input_name)

    for sample_id, sequence in file_iterator_tqdm:
        samples = add_id_to_samples(samples=samples, sample_id=sample_id)

        sequence_split = split_func(sequence)
        sequence_encoded = np.array(encode_func(sequence_split))

        samples[sample_id].inputs[input_name] = sequence_encoded

    return samples


def get_csv_id_sequence_iterator(
    data_source: str, ids_to_keep: Optional[Set[str]] = None
) -> Iterator[Tuple[str, str]]:
    df = pd.read_csv(data_source, index_col="ID", dtype={"ID": str})

    if "Sequence" not in df.columns:
        raise ValueError(
            f"Expected to find a 'Sequence' column in {data_source}, but didn't."
        )

    if ids_to_keep:
        df = df[df.index.isin(set(ids_to_keep))]

    for row in df.itertuples():
        cur_seq = row.Sequence
        if pd.isna(cur_seq):
            cur_seq = ""

        cur_index = row.Index
        assert isinstance(cur_index, str)
        assert isinstance(cur_seq, str)

        yield cur_index, cur_seq



from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Callable

import numpy as np
import torch
from aislib.misc_utils import ensure_path_exists
from matplotlib import pyplot as plt
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE
from torch import nn
from umap import UMAP

if TYPE_CHECKING:
    from eir.train_utils.evaluation import EvaluationResults


@dataclass
class LatentHookOutput:
    name: str
    outputs: np.ndarray
    ids: list[str]


def latent_analysis_wrapper(
    latent_outputs: LatentHookOutput,
    run_folder: Path,
    iteration: int | str,
) -> None:
    latent_output_folder = (
        run_folder / "latents/latent_outputs" / f"{iteration}" / latent_outputs.name
    )
    ensure_path_exists(path=latent_output_folder, is_folder=True)

    save_latent_outputs(
        outputs=latent_outputs.outputs,
        ids=latent_outputs.ids,
        folder=latent_output_folder,
    )

    outputs_parsed = parse_latent_outputs(latent_outputs=latent_outputs.outputs)
    plot_pca(outputs=outputs_parsed, folder=latent_output_folder)
    plot_tsne(outputs=outputs_parsed, folder=latent_output_folder)
    plot_umap(outputs=outputs_parsed, folder=latent_output_folder)


def register_latent_hook(
    model: nn.Module,
    layer_path: str,
) -> Callable[["EvaluationResults"], LatentHookOutput]:
    outputs = []

    def hook(
        module: nn.Module,
        input: tuple[torch.Tensor, ...],
        output: torch.Tensor,
    ) -> None:
        outputs.append(output.data.cpu().numpy())

    layer = model
    for name in layer_path.split("."):
        layer = getattr(layer, name)

    if not isinstance(layer, nn.Module):
        raise ValueError(f"Layer {layer_path} not found.")

    handle = layer.register_forward_hook(hook=hook)

    def get_outputs_and_remove_hook(
        evaluation_results: "EvaluationResults",
    ) -> LatentHookOutput:
        handle.remove()
        outputs_arr = np.concatenate(outputs, axis=0)
        ids = evaluation_results.all_ids
        latent_output_object = LatentHookOutput(
            outputs=outputs_arr,
            ids=ids,
            name=layer_path,
        )
        return latent_output_object

    return get_outputs_and_remove_hook


def parse_latent_outputs(latent_outputs: np.ndarray) -> np.ndarray:
    return latent_outputs.reshape(latent_outputs.shape[0], -1)


def save_latent_outputs(
    outputs: np.ndarray, ids: list[str], folder: Path
) -> np.ndarray:
    assert len(outputs) == len(ids), "Number of outputs and IDs must match."

    ids_array = np.array(ids)
    max_str_len = max(len(id_) for id_ in ids_array)

    structured_array = np.empty(
        len(outputs),
        dtype=[
            ("Latent", float, outputs.shape[1:]),
            ("ID", f"U{max_str_len}"),
        ],
    )
    structured_array["Latent"] = outputs
    structured_array["ID"] = ids_array

    np.save(file=str(folder / "latents.npy"), arr=structured_array)
    return structured_array


def plot_pca(outputs: np.ndarray, folder: Path) -> None:
    pca = PCA(n_components=2)
    reduced = pca.fit_transform(X=outputs)

    plt.figure(figsize=(8, 6))
    plt.scatter(reduced[:, 0], reduced[:, 1], s=2)

    variance_explained = pca.explained_variance_ratio_

    plt.xlabel(f"PCA 1 ({variance_explained[0]:.2f})")
    plt.ylabel(f"PCA 2 ({variance_explained[1]:.2f})")

    plt.title("PCA plot of latent space")

    plt.savefig(folder / "pca.png")


def plot_tsne(outputs: np.ndarray, folder: Path) -> None:
    tsne = TSNE(n_components=2)
    reduced = tsne.fit_transform(X=outputs)

    plt.figure(figsize=(8, 6))
    plt.scatter(reduced[:, 0], reduced[:, 1], s=2)

    plt.xlabel("t-SNE 1")
    plt.ylabel("t-SNE 2")

    plt.title("t-SNE plot of latent space")

    plt.savefig(folder / "tsne.png")


def plot_umap(outputs: np.ndarray, folder: Path) -> None:
    umap = UMAP(n_components=2)

    reduced = umap.fit_transform(X=outputs)

    plt.figure(figsize=(8, 6))
    plt.scatter(reduced[:, 0], reduced[:, 1], s=2)

    plt.xlabel("UMAP 1")
    plt.ylabel("UMAP 2")
    plt.title("UMAP plot of latent space")

    plt.savefig(folder / "umap.png")



from pathlib import Path
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np
import torch
from aislib.misc_utils import ensure_path_exists
from captum.attr._utils.visualization import visualize_image_attr_multiple
from torchvision.transforms import Normalize

from eir.interpretation.interpretation_utils import (
    get_basic_sample_attributions_to_analyse_generator,
    get_target_class_name,
)
from eir.setup.input_setup_modules.setup_image import ComputedImageInputInfo
from eir.setup.output_setup_modules.tabular_output_setup import (
    ComputedTabularOutputInfo,
)
from eir.setup.schemas import BasicInterpretationConfig

if TYPE_CHECKING:
    from eir.interpretation.interpretation import SampleAttribution
    from eir.setup.input_setup_modules.setup_image import ImageNormalizationStats
    from eir.train import Experiment


def analyze_image_input_attributions(
    experiment: "Experiment",
    input_name: str,
    target_column_name: str,
    output_name: str,
    target_column_type: str,
    attribution_outfolder: Path,
    all_attributions: list["SampleAttribution"],
) -> None:
    exp = experiment

    output_object = exp.outputs[output_name]
    assert isinstance(output_object, ComputedTabularOutputInfo)

    target_transformer = output_object.target_transformers[target_column_name]

    input_object = exp.inputs[input_name]
    assert isinstance(input_object, ComputedImageInputInfo)

    interpretation_config = input_object.input_config.interpretation_config
    assert isinstance(interpretation_config, BasicInterpretationConfig)

    samples_to_act_analyze_gen = get_basic_sample_attributions_to_analyse_generator(
        interpretation_config=interpretation_config, all_attributions=all_attributions
    )

    for sample_attribution in samples_to_act_analyze_gen:
        sample_target_labels = sample_attribution.sample_info.target_labels

        cur_label_name = get_target_class_name(
            sample_label=sample_target_labels[output_name][target_column_name],
            target_transformer=target_transformer,
            column_type=target_column_type,
            target_column_name=target_column_name,
        )

        attributions = sample_attribution.sample_attributions[input_name].squeeze(0)
        raw_input = sample_attribution.raw_inputs[input_name].cpu().numpy().squeeze(0)
        raw_input = un_normalize_image(
            normalized_img=raw_input,
            normalization_stats=input_object.normalization_stats,
        )
        raw_input = raw_input.transpose(1, 2, 0)
        attributions = attributions.transpose(1, 2, 0)

        figure, _ = visualize_image_attr_multiple(
            attr=attributions,
            original_image=raw_input,
            methods=["original_image", "heat_map"],
            signs=["all", "absolute_value"],
            show_colorbar=True,
            use_pyplot=False,
        )

        name = f"image_{sample_attribution.sample_info.ids[0]}_{cur_label_name}.pdf"
        output_path = attribution_outfolder / "single_samples" / name
        ensure_path_exists(path=output_path)
        figure.savefig(output_path, dpi=300)
        plt.close("all")


def un_normalize_image(
    normalized_img: np.ndarray, normalization_stats: "ImageNormalizationStats"
) -> np.ndarray:
    """
    Clip because sometimes we get values like 1.0000001 which will cause some libraries
    to do =/ 255, resulting in a black image.
    """
    means = torch.Tensor(normalization_stats.means)
    stds = torch.Tensor(normalization_stats.stds)
    un_norm = Normalize(
        mean=(-means / stds).tolist(),
        std=(1.0 / stds).tolist(),
    )

    img_as_tensor = torch.Tensor(normalized_img)
    img_un_normalized = un_norm(img_as_tensor).cpu().numpy()
    img_final = img_un_normalized.clip(0.0, 1.0)

    return img_final



from dataclasses import dataclass

import numpy as np
import torch
from torch import nn
from torch.nn import functional as F


@dataclass
class DiffusionConfig:
    time_steps: int
    betas: torch.Tensor
    sqrt_alphas_cumprod: torch.Tensor
    sqrt_one_minus_alphas_cumprod: torch.Tensor
    sqrt_recip_alphas: torch.Tensor
    posterior_variance: torch.Tensor


def initialize_diffusion_config(time_steps: int) -> DiffusionConfig:
    beta_start = 0.0001
    beta_end = 0.02
    betas = torch.linspace(beta_start, beta_end, time_steps)

    alphas = 1.0 - betas
    alphas_cumprod = torch.cumprod(alphas, dim=0)
    alphas_cumprod_prev = F.pad(alphas_cumprod[:-1], (1, 0), value=1.0)

    sqrt_recip_alphas = torch.sqrt(1.0 / alphas)
    sqrt_alphas_cumprod = torch.sqrt(alphas_cumprod)
    sqrt_one_minus_alphas_cumprod = torch.sqrt(1.0 - alphas_cumprod)
    posterior_variance = betas * (1.0 - alphas_cumprod_prev) / (1.0 - alphas_cumprod)

    return DiffusionConfig(
        time_steps=time_steps,
        betas=betas,
        sqrt_alphas_cumprod=sqrt_alphas_cumprod,
        sqrt_one_minus_alphas_cumprod=sqrt_one_minus_alphas_cumprod,
        sqrt_recip_alphas=sqrt_recip_alphas,
        posterior_variance=posterior_variance,
    )


def prepare_diffusion_batch(
    diffusion_config: DiffusionConfig,
    inputs: torch.Tensor,
    batch_size: int,
    num_steps: int,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Later we can return t from here as well if used in models.
    """
    t = torch.randint(
        low=0,
        high=num_steps,
        size=(batch_size,),
        device=inputs.device,
        dtype=torch.long,
    )

    noise = torch.randn_like(inputs)

    x_noisy = q_sample(
        config=diffusion_config,
        x_start=inputs,
        t=t,
        noise=noise,
    )

    return x_noisy, noise, t


def q_sample(
    config: DiffusionConfig,
    x_start: torch.Tensor,
    t: torch.Tensor,
    noise: torch.Tensor,
    input_scale: float = 1.0,
) -> torch.Tensor:
    """
    This and other functions adapted from
    https://huggingface.co/blog/annotated-diffusion
    """

    sqrt_alphas_cumprod_t = extract(
        a=config.sqrt_alphas_cumprod,
        t=t,
        x_shape=x_start.shape,
    )

    sqrt_one_minus_alphas_cumprod_t = extract(
        a=config.sqrt_one_minus_alphas_cumprod,
        t=t,
        x_shape=x_start.shape,
    )

    scaled_x_start = input_scale * x_start

    x_noisy = (
        sqrt_alphas_cumprod_t * scaled_x_start + sqrt_one_minus_alphas_cumprod_t * noise
    )

    return x_noisy


@torch.no_grad()
def p_sample_loop(
    config: DiffusionConfig,
    batch_inputs: dict,
    output_name: str,
    model: nn.Module,
    output_shape: tuple,
    time_steps: int,
) -> np.ndarray:
    device = next(model.parameters()).device

    current_state: torch.Tensor = torch.randn(output_shape, device=device)
    batch_inputs[output_name] = current_state

    batch_size = output_shape[0]

    for i in reversed(range(0, time_steps)):
        t = torch.full(
            size=(batch_size,),
            fill_value=i,
            device=device,
            dtype=torch.long,
        )

        current_state = p_sample(
            config=config,
            model=model,
            output_name=output_name,
            batch_inputs=batch_inputs,
            t=t,
            t_index=i,
        )

        batch_inputs[output_name] = current_state

    return current_state.cpu().numpy()


@torch.no_grad()
def p_sample(
    config: DiffusionConfig,
    model: nn.Module,
    batch_inputs: dict[str, torch.Tensor],
    output_name: str,
    t: torch.Tensor,
    t_index: int,
) -> torch.Tensor:

    current_state = batch_inputs[output_name]
    output_module = getattr(model.output_modules, output_name)
    t_emb = output_module.feature_extractor.timestep_embeddings(t)
    batch_inputs[f"__extras_{output_name}"] = t_emb

    betas_t = extract(
        a=config.betas,
        t=t,
        x_shape=current_state.shape,
    )
    sqrt_one_minus_alphas_cumprod_t = extract(
        a=config.sqrt_one_minus_alphas_cumprod,
        t=t,
        x_shape=current_state.shape,
    )
    sqrt_recip_alphas_t = extract(
        a=config.sqrt_recip_alphas,
        t=t,
        x_shape=current_state.shape,
    )

    model_outputs = model(batch_inputs)
    model_diffusion_output = model_outputs[output_name][output_name]

    model_mean = sqrt_recip_alphas_t * (
        current_state
        - betas_t * model_diffusion_output / sqrt_one_minus_alphas_cumprod_t
    )

    if t_index == 0:
        return model_mean
    else:
        posterior_variance_t = extract(
            a=config.posterior_variance,
            t=t,
            x_shape=current_state.shape,
        )
        noise = torch.randn_like(current_state)
        return model_mean + torch.sqrt(posterior_variance_t) * noise


def extract(
    a: torch.Tensor,
    t: torch.Tensor,
    x_shape: tuple,
) -> torch.Tensor:
    """
    Allows us to extract the appropriate index for a batch of indices.
    """
    batch_size = t.shape[0]
    out = a.gather(-1, t.cpu())
    return out.reshape(batch_size, *((1,) * (len(x_shape) - 1))).to(t.device)



from dataclasses import dataclass
from typing import TYPE_CHECKING, Dict

import torch
from torch import nn

if TYPE_CHECKING:
    from eir.setup.output_setup_modules.tabular_output_setup import (
        al_num_outputs_per_target,
    )


@dataclass
class LinearOutputModuleConfig:
    pass


class LinearOutputModule(nn.Module):
    def __init__(
        self,
        model_config: LinearOutputModuleConfig,
        input_dimension: int,
        num_outputs_per_target: "al_num_outputs_per_target",
    ):
        super().__init__()

        self.model_config = model_config
        self.input_dimension = input_dimension
        self.num_outputs_per_target = num_outputs_per_target
        self.total_outputs = sum(num_outputs_per_target.values())
        self.linear_layer = nn.Linear(
            in_features=input_dimension,
            out_features=self.total_outputs,
        )

        sorted_targets = sorted(num_outputs_per_target.items())
        target_names, target_sizes = zip(*sorted_targets)

        self.target_names = target_names
        self.target_sizes = list(target_sizes)

    def forward(self, inputs: torch.Tensor) -> Dict[str, torch.Tensor]:
        outputs = self.linear_layer(inputs)
        split_outputs = torch.split(outputs, self.target_sizes, dim=1)
        return {
            target: output for target, output in zip(self.target_names, split_outputs)
        }


def _get_linear_multi_task_branches(
    input_dimension: int,
    num_outputs_per_target: "al_num_outputs_per_target",
) -> nn.ModuleDict:
    multi_task_branches = nn.ModuleDict(
        {
            target: nn.Linear(
                in_features=input_dimension,
                out_features=num_outputs,
            )
            for target, num_outputs in num_outputs_per_target.items()
        }
    )

    return multi_task_branches



import os
from collections import OrderedDict
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from typing import (
    Any,
    Callable,
    Generator,
    Iterator,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    Union,
    cast,
    overload,
)

import pandas as pd
from tokenizers import Tokenizer
from tokenizers.models import BPE
from tokenizers.trainers import BpeTrainer
from tqdm import tqdm
from transformers import AutoTokenizer, PreTrainedTokenizer
from transformers.tokenization_utils_base import (
    EncodedInput,
    PreTokenizedInput,
    TextInput,
)

from eir.data_load.data_source_modules.deeplake_ops import (
    get_deeplake_input_source_iterable,
    is_deeplake_dataset,
    load_deeplake_dataset,
)
from eir.models.input.sequence.transformer_models import SequenceModelConfig
from eir.setup import schemas
from eir.setup.input_setup_modules.common import get_default_sequence_specials
from eir.setup.input_setup_modules.torchtext_port.utils import (
    get_tokenizer as get_pytorch_tokenizer,
)
from eir.setup.input_setup_modules.torchtext_port.vocab import Vocab
from eir.setup.input_setup_modules.torchtext_port.vocab_factory import (
    build_vocab_from_iterator,
)
from eir.setup.input_setup_modules.torchtext_port.vocab_factory import (
    vocab as pytorch_vocab_builder,
)
from eir.setup.schemas import al_tokenizer_choices
from eir.utils.logging import get_logger


class TokenizerProtocolRaw(Protocol):
    def __call__(self, raw_input: str) -> Sequence[str]: ...

    __closure__: Optional[Tuple[Any, ...]]


class TokenizerProtocolPreSplit(Protocol):
    def __call__(self, raw_input_split: Sequence[str]) -> Sequence[str]: ...

    __closure__: Optional[Tuple[Any, ...]]


class EncodeFuncProtocol(Protocol):
    def __call__(self, raw_input: Sequence[str] | str) -> Sequence[int]: ...


al_hf_tokenizer_inputs = Union[TextInput, PreTokenizedInput, EncodedInput]
al_sequence_input_objects_basic = Tuple[
    Vocab,
    "GatheredSequenceStats",
    TokenizerProtocolPreSplit | TokenizerProtocolRaw,
    EncodeFuncProtocol,
]
al_hf_encode_func = Callable[[al_hf_tokenizer_inputs], Sequence[int]]
al_sequence_input_objects_hf = Tuple[
    Vocab,
    "GatheredSequenceStats",
    PreTrainedTokenizer,
    al_hf_encode_func,
]

al_tokenizers = TokenizerProtocolRaw | TokenizerProtocolPreSplit
al_encode_funcs = EncodeFuncProtocol | al_hf_encode_func


logger = get_logger(name=__name__)


@dataclass
class ComputedSequenceInputInfo:
    input_config: schemas.InputConfig
    vocab: Vocab
    computed_max_length: int
    encode_func: al_encode_funcs
    tokenizer: Optional[al_tokenizers]


def set_up_sequence_input_for_training(
    input_config: schemas.InputConfig, *args, **kwargs
) -> ComputedSequenceInputInfo:
    model_config = input_config.model_config
    assert isinstance(model_config, SequenceModelConfig)

    sequence_input_object_func = _get_sequence_input_object_func(
        pretrained=model_config.pretrained_model
    )
    vocab, gathered_stats, tokenizer, encode_callable = sequence_input_object_func(
        input_config=input_config
    )

    input_type_info = input_config.input_type_info
    assert isinstance(input_type_info, schemas.SequenceInputDataConfig)

    gathered_stats = possibly_gather_all_stats_from_input(
        prev_gathered_stats=gathered_stats,
        input_source=input_config.input_info.input_source,
        vocab_file=input_type_info.vocab_file,
        split_on=input_type_info.split_on,
        max_length=input_type_info.max_length,
    )

    computed_max_length = get_max_length(
        max_length_config_value=input_type_info.max_length,
        gathered_stats=gathered_stats,
    )
    sequence_input_info = ComputedSequenceInputInfo(
        input_config=input_config,
        vocab=vocab,
        computed_max_length=computed_max_length,
        encode_func=encode_callable,
        tokenizer=tokenizer,
    )

    return sequence_input_info


class SequenceInputObjectGetterFunctionBasic(Protocol):
    def __call__(
        self, input_config: schemas.InputConfig
    ) -> al_sequence_input_objects_basic: ...


class SequenceInputObjectGetterFunctionHF(Protocol):
    def __call__(
        self, input_config: schemas.InputConfig
    ) -> al_sequence_input_objects_hf: ...


def _get_sequence_input_object_func(
    pretrained: bool,
) -> SequenceInputObjectGetterFunctionBasic | SequenceInputObjectGetterFunctionHF:
    if pretrained:
        return get_sequence_input_objects_from_pretrained
    else:
        return get_sequence_input_objects_from_input


def get_sequence_input_objects_from_input(
    input_config: schemas.InputConfig,
) -> al_sequence_input_objects_basic:
    gathered_stats = GatheredSequenceStats()
    input_type_info = input_config.input_type_info
    assert isinstance(input_type_info, schemas.SequenceInputDataConfig)

    tokenizer: TokenizerProtocolRaw | TokenizerProtocolPreSplit
    tokenizer, gathered_stats = get_tokenizer(
        input_config=input_config,
        gathered_stats=gathered_stats,
    )

    vocab = init_vocab(
        source=input_config.input_info.input_source,
        inner_key=input_config.input_info.input_inner_key,
        tokenizer_name=input_type_info.tokenizer,
        split_on=input_type_info.split_on,
        vocab_file=input_type_info.vocab_file,
        min_freq=input_type_info.min_freq,
        gathered_stats=gathered_stats,
        tokenizer=tokenizer,
    )

    encode_func: EncodeFuncProtocol
    encode_func = get_tokenizer_encode_func(
        tokenizer=tokenizer,
        pytorch_vocab=vocab,
    )

    return vocab, gathered_stats, tokenizer, encode_func


def init_vocab(
    source: str,
    inner_key: Optional[str],
    tokenizer_name: str,
    split_on: Optional[str],
    vocab_file: Optional[str],
    min_freq: int,
    gathered_stats: "GatheredSequenceStats",
    tokenizer: TokenizerProtocolRaw | TokenizerProtocolPreSplit,
) -> Vocab:
    if tokenizer_name == "bpe":
        assert gathered_stats.total_count > 0
        tokenizer_object = extract_tokenizer_object_from_function(
            tokenizer_callable=tokenizer
        )
        vocab = sync_hf_and_pytorch_vocab(hf_tokenizer=tokenizer_object)
    else:
        assert gathered_stats.total_count == 0

        vocab_iter = get_vocab_iterator(
            input_source=source,
            split_on=split_on,
            gathered_stats=gathered_stats,
            vocab_file=vocab_file,
            deeplake_inner_key=inner_key,
        )
        tokenized_vocab_iter = get_tokenized_vocab_iterator(
            vocab_iterator=vocab_iter, tokenizer=tokenizer
        )

        min_freq = _init_min_freq(
            vocab_file=vocab_file,
            min_freq=min_freq,
        )

        vocab = build_vocab_from_iterator(
            iterator=tokenized_vocab_iter,
            specials=get_default_sequence_specials(),
            min_freq=min_freq,
        )

    vocab.set_default_index(vocab["<unk>"])

    return vocab


def _init_min_freq(
    vocab_file: Optional[str],
    min_freq: int,
) -> int:
    if vocab_file:
        logger.info(
            "Minimum word/token frequency will be set to 0 as vocabulary is loaded "
            "from file %s.",
            vocab_file,
        )
        min_freq = 1
    return min_freq


def extract_tokenizer_object_from_function(
    tokenizer_callable: TokenizerProtocolRaw | TokenizerProtocolPreSplit,
) -> Tokenizer | PreTrainedTokenizer:
    closure = tokenizer_callable.__closure__
    assert closure is not None
    assert len(closure) == 1

    tokenizer_object = closure[0].cell_contents
    assert isinstance(tokenizer_object, (Tokenizer, PreTrainedTokenizer))
    return tokenizer_object


def get_tokenizer(
    input_config: schemas.InputConfig,
    gathered_stats: "GatheredSequenceStats",
) -> tuple[TokenizerProtocolPreSplit | TokenizerProtocolRaw, "GatheredSequenceStats"]:
    input_type_info = input_config.input_type_info
    assert isinstance(input_type_info, schemas.SequenceInputDataConfig)
    tokenizer_name = input_type_info.tokenizer

    tokenizer: TokenizerProtocolPreSplit | TokenizerProtocolRaw
    if tokenizer_name == "bpe":
        vocab_iter = get_vocab_iterator(
            input_source=input_config.input_info.input_source,
            split_on=input_type_info.split_on,
            gathered_stats=gathered_stats,
            vocab_file=input_type_info.vocab_file,
            deeplake_inner_key=input_config.input_info.input_inner_key,
        )

        vocab_file = input_type_info.vocab_file
        tokenizer = get_bpe_tokenizer(
            vocab_iterator=vocab_iter,
            vocab_file=vocab_file,
            vocab_size=input_type_info.adaptive_tokenizer_max_vocab_size,
            split_on=input_type_info.split_on,
        )

    else:
        tokenizer = get_basic_tokenizer(
            tokenizer_name=tokenizer_name,
            tokenizer_language=input_type_info.tokenizer_language,
        )

    return tokenizer, gathered_stats


def _get_default_specials_map() -> dict:
    mapping = {
        "bos_token": "<bos>",
        "unk_token": "<unk>",
        "mask_token": "<mask>",
        "pad_token": "<pad>",
        "eos_token": "<eos>",
    }

    default_specials = get_default_sequence_specials()
    assert set(mapping.values()) == set(default_specials)

    return mapping


def get_sequence_input_objects_from_pretrained(
    input_config: schemas.InputConfig,
) -> al_sequence_input_objects_hf:
    input_type_info = input_config.input_type_info
    assert isinstance(input_type_info, schemas.SequenceInputDataConfig)
    vocab_file = input_type_info.vocab_file
    if vocab_file:
        raise ValueError(
            "Using a vocabulary file not supported when using pre-trained models as "
            "their training vocabulary will be used."
        )

    gathered_stats = GatheredSequenceStats()
    hf_model_name = input_config.model_config.model_type
    hf_tokenizer = _get_hf_tokenizer(hf_model_name=hf_model_name)

    def _passthrough_hf_encode(
        raw_input_split: al_hf_tokenizer_inputs,
    ) -> Sequence[int]:
        return hf_tokenizer.encode(text=raw_input_split, is_split_into_words=True)

    vocab = sync_hf_and_pytorch_vocab(hf_tokenizer=hf_tokenizer)

    return vocab, gathered_stats, hf_tokenizer, _passthrough_hf_encode


def sync_hf_and_pytorch_vocab(hf_tokenizer: Tokenizer | PreTrainedTokenizer) -> Vocab:
    hf_tokenizer_vocab = hf_tokenizer.get_vocab()
    hf_tokenizer_vocab_sorted = OrderedDict(
        {k: v for k, v in sorted(hf_tokenizer_vocab.items(), key=lambda item: item[1])}
    )
    vocab = pytorch_vocab_builder(ordered_dict=hf_tokenizer_vocab_sorted, min_freq=0)

    return vocab


def _get_hf_tokenizer(hf_model_name: str) -> PreTrainedTokenizer:
    """
    See https://github.com/huggingface/transformers/issues/5486 for why we need to
    set the environment variable.
    """
    os.environ["TOKENIZERS_PARALLELISM"] = "false"

    hf_tokenizer = AutoTokenizer.from_pretrained(
        pretrained_model_name_or_path=hf_model_name, add_prefix_space=True
    )

    hf_tokenizer = _add_specials_to_hf_tokenizer(hf_tokenizer=hf_tokenizer)
    return hf_tokenizer


def _add_specials_to_hf_tokenizer(
    hf_tokenizer: PreTrainedTokenizer,
) -> PreTrainedTokenizer:
    hf_tokenizer_copy = deepcopy(hf_tokenizer)
    name_special_token_map = _get_default_specials_map()

    specials_tokens_to_add = {}
    for special_token_name, special_token in name_special_token_map.items():
        if special_token_name not in hf_tokenizer_copy.special_tokens_map:
            specials_tokens_to_add[special_token_name] = special_token

    hf_tokenizer_copy.add_special_tokens(special_tokens_dict=specials_tokens_to_add)
    logger.debug("Special tokens %s added to %s.", specials_tokens_to_add, hf_tokenizer)

    return hf_tokenizer_copy


def get_bpe_tokenizer(
    vocab_iterator: Optional[Iterator],
    vocab_file: Optional[str],
    vocab_size: Optional[int],
    split_on: Optional[str],
) -> TokenizerProtocolRaw | TokenizerProtocolPreSplit:
    tokenizer = _get_bpe_tokenizer_object(
        vocab_iterator=vocab_iterator,
        vocab_file=vocab_file,
        vocab_size=vocab_size,
    )

    def _tokenize_raw(raw_input: str) -> Sequence[str]:
        tokens = tokenizer.encode(
            sequence=raw_input,
            is_pretokenized=False,
        ).tokens
        return tokens

    def _tokenize_pre_split(raw_input_split: Sequence[str]) -> Sequence[str]:
        tokens = tokenizer.encode(
            sequence=raw_input_split,
            is_pretokenized=True,
        ).tokens
        return tokens

    if split_on is None:
        return cast(TokenizerProtocolRaw, _tokenize_raw)
    else:
        return cast(TokenizerProtocolPreSplit, _tokenize_pre_split)


def _get_bpe_tokenizer_object(
    vocab_iterator: Optional[Iterator],
    vocab_file: Optional[str],
    vocab_size: Optional[int],
) -> Tokenizer:
    if vocab_file:
        logger.info("Loading BPE vocabulary from file %s.", vocab_file)

        if not vocab_file.endswith(".json"):
            raise ValueError(
                "Vocabulary file must be a HuggingFace Tokenizers "
                "compatible .json file."
            )

        tokenizer = Tokenizer.from_file(path=vocab_file)
    else:
        assert vocab_iterator is not None

        logger.info("Training BPE tokenizer from source data.")

        tokenizer = Tokenizer(model=BPE(unk_token="<unk>"))

        special_tokens = get_default_sequence_specials()
        if vocab_size is not None:
            trainer = BpeTrainer(
                special_tokens=special_tokens,
                vocab_size=vocab_size,
            )
        else:
            trainer = BpeTrainer(
                special_tokens=special_tokens,
            )

        tokenizer.train_from_iterator(iterator=vocab_iterator, trainer=trainer)

    return tokenizer


def get_basic_tokenizer(
    tokenizer_name: al_tokenizer_choices,  # type: ignore
    tokenizer_language: Optional[str],
) -> TokenizerProtocolPreSplit:
    def _identity(raw_input_split: Sequence[str]) -> Sequence[str]:
        return raw_input_split

    if not tokenizer_name:
        return _identity

    _validate_pytorch_tokenizer_args(
        tokenizer_name=tokenizer_name, tokenizer_language=tokenizer_language
    )
    logger.debug(
        "Using tokenizer '%s' with language '%s'.", tokenizer_name, tokenizer_language
    )

    tokenizer = get_pytorch_tokenizer(
        tokenizer=tokenizer_name, language=tokenizer_language
    )

    def _join_and_tokenize(raw_input_split: Sequence[str]) -> Sequence[str]:
        """
        " ".join(...) since the PyTorch tokenizers operate on a single string.
        Can be optimized later by not splitting the input in the first place.
        """
        input_joined = " ".join(raw_input_split)
        return tokenizer(input_joined)

    return cast(TokenizerProtocolPreSplit, _join_and_tokenize)


def _validate_pytorch_tokenizer_args(
    tokenizer_name: al_tokenizer_choices,  # type: ignore
    tokenizer_language: Union[str, None],
) -> None:
    tokenizer_language_passed = tokenizer_language is not None
    tokenizer_does_not_support_language = tokenizer_name not in (
        "spacy",
        "basic_english",
    )

    if tokenizer_language_passed and tokenizer_does_not_support_language:
        raise ValueError(
            "Tokenizer '%s' does not support setting a language (got '%s'). "
            "Please leave it as None.",
            tokenizer_name,
            tokenizer_language,
        )


def get_tokenizer_encode_func(
    tokenizer: TokenizerProtocolRaw | TokenizerProtocolPreSplit,
    pytorch_vocab: Vocab,
) -> EncodeFuncProtocol:
    @overload
    def _encode_func(raw_input: str) -> Sequence[int]: ...

    @overload
    def _encode_func(raw_input: Sequence[str]) -> Sequence[int]: ...

    def _encode_func(raw_input):
        input_tokenized = tokenizer(raw_input)
        input_as_ids = pytorch_vocab(input_tokenized)
        return input_as_ids

    return _encode_func


def get_tokenized_vocab_iterator(
    vocab_iterator: Iterator[Sequence[str] | str],
    tokenizer: TokenizerProtocolRaw | TokenizerProtocolPreSplit,
) -> Generator[Sequence[str], None, None]:
    @overload
    def _do_tokenize(list_of_words_: str) -> Sequence[str]: ...

    @overload
    def _do_tokenize(list_of_words_: Sequence[str]) -> Sequence[str]: ...

    def _do_tokenize(list_of_words_):
        return tokenizer(list_of_words_)

    for list_of_words in vocab_iterator:
        yield _do_tokenize(list_of_words_=list_of_words)


def get_vocab_iterator(
    input_source: str,
    split_on: Optional[str],
    gathered_stats: "GatheredSequenceStats",
    vocab_file: Union[str, None] = None,
    deeplake_inner_key: Optional[str] = None,
) -> Generator[Sequence[str], None, None]:
    """
    Note: When using a vocabulary file, we explicitly expect one token per line,
    therefore we do not split on any character.
    """

    if vocab_file is None:
        logger.info(
            "Vocabulary will be collected from input source %s, "
            "splitting tokens on '%s'.",
            input_source,
            split_on,
        )
        vocab_iter = yield_tokens_from_source(
            data_source=input_source,
            split_on=split_on,
            gathered_stats=gathered_stats,
            deeplake_inner_key=deeplake_inner_key,
        )
    else:
        logger.info(
            "Vocabulary for %s will be collected from vocabulary file %s.",
            input_source,
            vocab_file,
        )
        vocab_iter = yield_tokens_from_file(
            file_path=vocab_file,
            split_on=None,
            gathered_stats=gathered_stats,
        )

    return vocab_iter


@dataclass
class GatheredSequenceStats:
    total_count: int = 0
    total_files: int = 0
    max_length: int = 0


def yield_tokens_from_source(
    data_source: str,
    split_on: Optional[str],
    gathered_stats: GatheredSequenceStats,
    deeplake_inner_key: Optional[str] = None,
):
    data_source_path = Path(data_source)

    if is_deeplake_dataset(data_source=str(data_source_path)):
        assert deeplake_inner_key is not None
        yield from yield_tokens_from_deeplake_dataset(
            data_source=data_source_path,
            split_on=split_on,
            gathered_stats=gathered_stats,
            inner_key=deeplake_inner_key,
        )

    elif data_source_path.is_dir():
        iterator = tqdm(Path(data_source).iterdir(), desc="Vocabulary Setup")
        for file in iterator:
            yield from yield_tokens_from_file(
                file_path=str(file), split_on=split_on, gathered_stats=gathered_stats
            )

    elif data_source_path.is_file():
        assert data_source_path.suffix == ".csv"
        yield from yield_tokens_from_csv(
            file_path=data_source, split_on=split_on, gathered_stats=gathered_stats
        )

    return gathered_stats


def yield_tokens_from_deeplake_dataset(
    data_source: Path,
    split_on: Optional[str],
    gathered_stats: GatheredSequenceStats,
    inner_key: str,
) -> Generator[Sequence[str], None, None]:
    deeplake_ds = load_deeplake_dataset(data_source=str(data_source))
    deeplake_iter = get_deeplake_input_source_iterable(
        deeplake_dataset=deeplake_ds, inner_key=inner_key
    )

    split_func = get_sequence_split_function(split_on=split_on)

    for sample in deeplake_iter:
        cur_sequence = sample.text()
        cur_line = split_func(cur_sequence)

        cur_length = len(cur_line)
        gathered_stats.total_count += len(cur_line)

        if cur_length > gathered_stats.max_length:
            gathered_stats.max_length = cur_length

        yield cur_line


def yield_tokens_from_file(
    file_path: str, split_on: Optional[str], gathered_stats: GatheredSequenceStats
) -> Generator[Sequence[str], None, None]:
    gathered_stats.total_files += 1

    split_func = get_sequence_split_function(split_on=split_on)

    with open(file_path, "r") as f:
        for line in f:
            line_parsed = line[:-1] if line.endswith("\n") else line
            cur_line = split_func(line_parsed)

            cur_length = len(cur_line)

            if split_on is None:
                gathered_stats.total_count += 1
            else:
                gathered_stats.total_count += len(cur_line)

            if cur_length > gathered_stats.max_length:
                gathered_stats.max_length = cur_length

            yield cur_line


def yield_tokens_from_csv(
    file_path: str, split_on: Optional[str], gathered_stats: GatheredSequenceStats
) -> Generator[Sequence[str], None, None]:
    split_func = get_sequence_split_function(split_on=split_on)

    df = pd.read_csv(filepath_or_buffer=file_path, index_col="ID", dtype={"ID": str})
    if "Sequence" not in df.columns:
        raise ValueError(
            "CSV file '%s' does not have a column named 'Sequence'. "
            "Please ensure that the column name is correct and present.",
            file_path,
        )

    iterator = tqdm(df.itertuples(), desc="Vocabulary Setup")
    for row in iterator:
        cur_sequence = row.Sequence
        if pd.isna(cur_sequence):
            cur_sequence = ""

        cur_line = split_func(cur_sequence)

        cur_length = len(cur_line)
        gathered_stats.total_count += len(cur_line)
        gathered_stats.total_files += 1

        if cur_length > gathered_stats.max_length:
            gathered_stats.max_length = cur_length

        yield cur_line


def get_sequence_split_function(
    split_on: Optional[str],
) -> Callable[[str], list[str] | str]:
    match split_on:
        case "":
            return lambda x: list(x)
        case None:
            return lambda x: x
        case _:
            return lambda x: x.split(split_on)


class ReturnSavingGenerator:
    def __init__(self, gen):
        self.gen = gen

    def __iter__(self):
        self.caught_return_value = yield from self.gen


def possibly_gather_all_stats_from_input(
    prev_gathered_stats: GatheredSequenceStats,
    input_source: str,
    vocab_file: Optional[str],
    split_on: Optional[str],
    max_length: schemas.al_max_sequence_length,
) -> GatheredSequenceStats:
    """
    Note that we use all(...) there to exhaust the generator object, so that the
    stats get accumulated in the GatheredSequenceStats().
    """
    gathered_stats = prev_gathered_stats

    dynamic_and_vocab_file = vocab_file and max_length in {"average", "max"}

    if dynamic_and_vocab_file:
        logger.info(
            "Doing a full pass over input data despite vocabulary file %s being "
            "present, as dynamic max length '%s' was requested.",
            vocab_file,
            max_length,
        )
        vocab_iter = yield_tokens_from_source(
            data_source=input_source,
            split_on=split_on,
            gathered_stats=GatheredSequenceStats(),
        )
        value_keeping_gen = ReturnSavingGenerator(gen=vocab_iter)
        all(value_keeping_gen)
        gathered_stats = value_keeping_gen.caught_return_value

    return gathered_stats


def get_max_length(
    max_length_config_value: schemas.al_max_sequence_length,
    gathered_stats: GatheredSequenceStats,
):
    if isinstance(max_length_config_value, int):
        return max_length_config_value

    if max_length_config_value == "max":
        logger.info(
            "Using inferred max length found in sequence data source as %d",
            gathered_stats.max_length,
        )
        return gathered_stats.max_length
    elif max_length_config_value == "average":
        average_length = gathered_stats.total_count // gathered_stats.total_files
        logger.info(
            "Using inferred average length found in sequence data source as %d",
            average_length,
        )
        return average_length

    raise ValueError("Unknown max length config value %s.", max_length_config_value)



from typing import TYPE_CHECKING, DefaultDict, Optional

import pandas as pd
from tqdm import tqdm

from eir.data_load.data_source_modules.common_utils import add_id_to_samples

if TYPE_CHECKING:
    from eir.data_load.data_utils import Sample


def add_tabular_data_to_samples(
    df_tabular: pd.DataFrame,
    samples: DefaultDict[str, "Sample"],
    ids_to_keep: Optional[set[str]] = None,
    source_name: str = "Tabular Data",
) -> DefaultDict[str, "Sample"]:

    def _get_tabular_iterator(ids_to_keep_: Optional[set[str]] = None):
        fields = df_tabular.columns
        for row in df_tabular.itertuples():
            sample_id_ = str(getattr(row, "Index"))
            if ids_to_keep_ is not None and sample_id_ not in ids_to_keep_:
                continue

            tabular_inputs_ = {field: getattr(row, field) for field in fields}

            yield sample_id_, tabular_inputs_

    known_length = None if ids_to_keep is None else len(ids_to_keep)
    tabular_iterator = tqdm(
        _get_tabular_iterator(ids_to_keep_=ids_to_keep),
        desc=source_name,
        total=known_length,
    )

    for sample_id, tabular_inputs in tabular_iterator:
        samples = add_id_to_samples(samples=samples, sample_id=sample_id)

        samples[sample_id].inputs[source_name] = tabular_inputs

    return samples



from copy import copy
from dataclasses import dataclass
from functools import partial
from typing import (
    TYPE_CHECKING,
    Callable,
    List,
    Literal,
    Optional,
    Protocol,
    Sequence,
    Union,
)

import torch
from torch import nn

from eir.models.input.sequence.transformer_models import PositionalEmbedding
from eir.models.layers.attention_layers import LinearAttention
from eir.models.layers.lcl_layers import LCL, LCLResidualBlock
from eir.models.layers.norm_layers import LayerScale
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.models.output.array.array_output_modules import LCLOutputModelConfig
    from eir.setup.input_setup_modules.common import DataDimensions

logger = get_logger(__name__)


class FlattenFunc(Protocol):
    def __call__(
        self,
        x: torch.Tensor,
    ) -> torch.Tensor: ...


@dataclass
class SimpleLCLModelConfig:
    """
    :param fc_repr_dim:
        Controls the number of output sets in the first and only split layer. Analogous
        to channels in CNNs.
    :param num_lcl_chunks:
        Controls the number of splits applied to the input. E.g. with a input with of
        800, using ``num_lcl_chunks=100`` will result in a kernel width of 8,
        meaning 8 elements in the flattened input. If using a SNP inputs with a one-hot
        encoding of 4 possible values, this will result in 8/2 = 2 SNPs per locally
        connected area.
    :param l1:
        L1 regularization applied to the first and only locally connected layer.
    """

    fc_repr_dim: int = 12
    num_lcl_chunks: int = 64
    l1: float = 0.00


class SimpleLCLModel(nn.Module):
    def __init__(
        self,
        model_config: SimpleLCLModelConfig,
        data_dimensions: "DataDimensions",
        flatten_fn: FlattenFunc,
    ):
        super().__init__()

        self.model_config = model_config
        self.data_dimensions = data_dimensions
        self.flatten_fn = flatten_fn

        num_chunks = self.model_config.num_lcl_chunks
        self.fc_0 = LCL(
            in_features=self.fc_1_in_features,
            out_feature_sets=self.model_config.fc_repr_dim,
            num_chunks=num_chunks,
            bias=True,
        )

        self._init_weights()

    @property
    def fc_1_in_features(self) -> int:
        return self.data_dimensions.num_elements()

    @property
    def l1_penalized_weights(self) -> torch.Tensor:
        return self.fc_0.weight

    @property
    def num_out_features(self) -> int:
        return self.fc_0.out_features

    def _init_weights(self):
        pass

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        out = self.flatten_fn(x=input)

        out = self.fc_0(out)

        return out


@dataclass
class LCLModelConfig:
    """
    Note that when using the automatic network setup, kernel widths will get expanded
    to ensure that the feature representations become smaller as they are propagated
    through the network.

    :param patch_size:
        Controls the size of the patches used in the first layer. If set to ``None``,
        the input is flattened according to the torch ``flatten`` function. Note that
        when using this parameter, we generally want the kernel width to be set to
        the multiplication of the patch size. Order follows PyTorch convention, i.e.,
        [channels, height, width].

    :param layers:
        Controls the number of layers in the model. If set to ``None``, the model will
        automatically set up the number of layers according to the ``cutoff`` parameter
        value.

    :param kernel_width:
        With of the locally connected kernels. Note that in the context of genomic
        inputs this refers to the flattened input,
        meaning that if we have a one-hot encoding of 4 values (e.g. SNPs), 12
        refers to 12/4 = 3 SNPs per locally connected window. Can be set to ``None`` if
        the ``num_lcl_chunks`` parameter is set, which means that the kernel width
        will be set automatically according to

    :param first_kernel_expansion:
        Factor to extend the first kernel. This value can both be positive or negative.
        For example in the case of ``kernel_width=12``, setting
        ``first_kernel_expansion=2`` means that the first kernel will have a width of
        24, whereas other kernels will have a width of 12. When using a negative value,
        divides the first kernel by the value instead of multiplying.

    :param channel_exp_base:
        Which power of 2 to use in order to set the number of channels/weight sets in
        the network. For example, setting ``channel_exp_base=3`` means that 2**3=8
        weight sets will be used.

    :param first_channel_expansion:
        Whether to expand / shrink the number of channels in the first layer as compared
        to other layers in the network. Works analogously to the
        ``first_kernel_expansion`` parameter.

    :param num_lcl_chunks:
        Controls the number of splits applied to the input. E.g. with a input width of
        800, using ``num_lcl_chunks=100`` will result in a kernel width of 8,
        meaning 8 elements in the flattened input. If using a SNP inputs with a one-hot
        encoding of 4 possible values, this will result in 8/2 = 2 SNPs per locally
        connected area.

    :param rb_do:
        Dropout in the residual blocks.

    :param stochastic_depth_p:
        Probability of dropping input.

    :param l1:
        L1 regularization applied to the first layer in the network.

    :param cutoff:
        Feature dimension cutoff where the automatic network setup stops adding layers.
        The 'auto' option is only supported when using the model for array *outputs*,
        and will set the cutoff to roughly the number of output features.

    :param direction:
        Whether to use a "down" or "up" network. "Down" means that the feature
        representation will get smaller as it is propagated through the network, whereas
        "up" means that the feature representation will get larger.

    :param attention_inclusion_cutoff:
        Cutoff to start including attention blocks in the network. If set to ``None``,
        no attention blocks will be included. The cutoff here refers to the "length"
        dimension of the input after reshaping according to the output_feature_sets
        in the preceding layer. For example, if we 1024 output features, and we have
        4 output feature sets, the length dimension will be 1024/4 = 256. With an
        attention cutoff >= 256, the attention block will be included.
    """

    patch_size: Optional[tuple[int, int, int]] = None

    layers: Union[None, List[int]] = None

    kernel_width: int | Literal["patch"] = 16
    first_kernel_expansion: int = -2

    channel_exp_base: int = 2
    first_channel_expansion: int = 1

    num_lcl_chunks: Union[None, int] = None

    rb_do: float = 0.10
    stochastic_depth_p: float = 0.00
    l1: float = 0.00

    cutoff: int | Literal["auto"] = 1024
    direction: Literal["down", "up"] = "down"
    attention_inclusion_cutoff: Optional[int] = None


class LCLModel(nn.Module):
    def __init__(
        self,
        model_config: Union[LCLModelConfig, "LCLOutputModelConfig"],
        data_dimensions: "DataDimensions",
        flatten_fn: FlattenFunc,
        dynamic_cutoff: Optional[int] = None,
    ):
        super().__init__()

        self.model_config = model_config
        self.data_dimensions = data_dimensions
        self.flatten_fn = flatten_fn

        kernel_width = parse_kernel_width(
            kernel_width=self.model_config.kernel_width,
            patch_size=self.model_config.patch_size,
        )

        fc_0_kernel_size = calc_value_after_expansion(
            base=kernel_width,
            expansion=self.model_config.first_kernel_expansion,
        )
        fc_0_out_feature_sets = calc_value_after_expansion(
            base=2**self.model_config.channel_exp_base,
            expansion=self.model_config.first_channel_expansion,
        )
        self.fc_0 = LCL(
            in_features=self.fc_1_in_features,
            out_feature_sets=fc_0_out_feature_sets,
            kernel_size=fc_0_kernel_size,
            bias=True,
        )

        cutoff = dynamic_cutoff or self.model_config.cutoff
        assert isinstance(cutoff, int)

        lcl_parameter_spec = LCParameterSpec(
            in_features=self.fc_0.out_features,
            kernel_width=kernel_width,
            channel_exp_base=self.model_config.channel_exp_base,
            dropout_p=self.model_config.rb_do,
            cutoff=cutoff,
            stochastic_depth_p=self.model_config.stochastic_depth_p,
            attention_inclusion_cutoff=self.model_config.attention_inclusion_cutoff,
            direction=self.model_config.direction,
        )
        self.lcl_blocks = _get_lcl_blocks(
            lcl_spec=lcl_parameter_spec,
            block_layer_spec=self.model_config.layers,
        )

        self.output_shape = (1, 1, self.lcl_blocks[-1].out_features)

        self._init_weights()

    @property
    def fc_1_in_features(self) -> int:
        return self.data_dimensions.num_elements()

    @property
    def l1_penalized_weights(self) -> torch.Tensor:
        return self.fc_0.weight

    @property
    def num_out_features(self) -> int:
        return self.lcl_blocks[-1].out_features

    def _init_weights(self):
        pass

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        out = self.flatten_fn(x=input)

        out = self.fc_0(out)
        out = self.lcl_blocks(out)

        return out


def parse_kernel_width(
    kernel_width: int | Literal["patch"],
    patch_size: Optional[tuple[int, int, int]],
) -> int:
    if kernel_width == "patch":
        if patch_size is None:
            raise ValueError(
                "kernel_width set to 'patch', but no patch_size was specified."
            )
        kernel_width = patch_size[0] * patch_size[1] * patch_size[2]
    return kernel_width


def flatten_h_w_fortran(x: torch.Tensor) -> torch.Tensor:
    """
    This is needed when e.g. flattening one-hot inputs that are ordered in a columns
    wise fashion (meaning that each column is a one-hot feature),
    and we want to make sure the first part of the flattened tensor is the first column,
    i.e. first one-hot element.
    """
    column_order_flattened = x.transpose(2, 3).flatten(1)
    return column_order_flattened


def calc_value_after_expansion(base: int, expansion: int, min_value: int = 0) -> int:
    if expansion > 0:
        return base * expansion
    elif expansion < 0:
        abs_expansion = abs(expansion)
        return max(min_value, base // abs_expansion)
    return base


@dataclass
class LCParameterSpec:
    in_features: int
    kernel_width: int
    channel_exp_base: int
    dropout_p: float
    stochastic_depth_p: float
    cutoff: int
    attention_inclusion_cutoff: Optional[int] = None
    direction: Literal["down", "up"] = "down"


def _get_lcl_blocks(
    lcl_spec: LCParameterSpec,
    block_layer_spec: Optional[Sequence[int]],
) -> nn.Sequential:
    factory = _get_lcl_block_factory(block_layer_spec=block_layer_spec)

    blocks = factory(lcl_spec)

    return blocks


def _get_lcl_block_factory(
    block_layer_spec: Optional[Sequence[int]],
) -> Callable[[LCParameterSpec], nn.Sequential]:
    if not block_layer_spec:
        return generate_lcl_residual_blocks_auto

    auto_factory = partial(
        _generate_lcl_blocks_from_spec,
        block_layer_spec=block_layer_spec,
    )

    return auto_factory


def _generate_lcl_blocks_from_spec(
    lcl_parameter_spec: LCParameterSpec,
    block_layer_spec: Sequence[int],
) -> nn.Sequential:
    s = lcl_parameter_spec
    block_layer_spec_copy: list[int] = list(copy(block_layer_spec))

    first_block = LCLResidualBlock(
        in_features=s.in_features,
        kernel_size=s.kernel_width,
        out_feature_sets=2**s.channel_exp_base,
        dropout_p=s.dropout_p,
        full_preactivation=True,
    )

    block_modules = [first_block]
    block_layer_spec_copy[0] -= 1

    for cur_layer_index, block_dim in enumerate(block_layer_spec_copy):
        for block in range(block_dim):
            cur_out_feature_sets = 2 ** (s.channel_exp_base + cur_layer_index)
            cur_kernel_width = s.kernel_width

            cur_out_feature_sets, cur_kernel_width = _adjust_auto_params(
                cur_out_feature_sets=cur_out_feature_sets,
                cur_kernel_width=cur_kernel_width,
                direction=s.direction,
            )

            cur_size = block_modules[-1].out_features

            cur_block = LCLResidualBlock(
                in_features=cur_size,
                kernel_size=cur_kernel_width,
                out_feature_sets=cur_out_feature_sets,
                dropout_p=s.dropout_p,
                stochastic_depth_p=s.stochastic_depth_p,
            )

            block_modules.append(cur_block)

    return nn.Sequential(*block_modules)


def _adjust_auto_params(
    cur_out_feature_sets: int, cur_kernel_width: int, direction: Literal["down", "up"]
) -> tuple[int, int]:
    """
    Down: increase kernel width until it is larger than the number of output feature
    sets.
    Up: increase number of output feature sets until it is larger than the kernel width.

    """
    if direction == "down":
        while cur_out_feature_sets >= cur_kernel_width:
            cur_kernel_width *= 2
    elif direction == "up":
        while cur_out_feature_sets <= cur_kernel_width:
            cur_out_feature_sets *= 2
    else:
        raise ValueError(f"Unknown direction: {direction}")

    return cur_out_feature_sets, cur_kernel_width


def generate_lcl_residual_blocks_auto(lcl_parameter_spec: LCParameterSpec):
    """
    TODO:   Create some over-engineered abstraction for this and
            ``_generate_lcl_blocks_from_spec`` if feeling bored.
    """

    s = lcl_parameter_spec

    first_block = LCLResidualBlock(
        in_features=s.in_features,
        kernel_size=s.kernel_width,
        out_feature_sets=2**s.channel_exp_base,
        dropout_p=s.dropout_p,
        full_preactivation=True,
    )

    block_modules: list[LCLResidualBlock | LCLAttentionBlock]
    block_modules = [first_block]

    if _do_add_attention(
        attention_inclusion_cutoff=s.attention_inclusion_cutoff,
        in_features=first_block.out_features,
        embedding_dim=first_block.out_feature_sets,
    ):
        cur_attention_block = LCLAttentionBlock(
            embedding_dim=first_block.out_feature_sets,
            in_features=first_block.out_features,
        )
        block_modules.append(cur_attention_block)

    while True:
        cur_no_blocks = len(block_modules)
        cur_index = cur_no_blocks // 2

        cur_out_feature_sets = 2 ** (s.channel_exp_base + cur_index)
        cur_kernel_width = s.kernel_width
        cur_out_feature_sets, cur_kernel_width = _adjust_auto_params(
            cur_out_feature_sets=cur_out_feature_sets,
            cur_kernel_width=cur_kernel_width,
            direction=s.direction,
        )

        cur_size = block_modules[-1].out_features

        if _should_break_auto(
            cur_size=cur_size,
            cutoff=s.cutoff,
            direction=s.direction,
        ):
            break

        cur_block = LCLResidualBlock(
            in_features=cur_size,
            kernel_size=cur_kernel_width,
            out_feature_sets=cur_out_feature_sets,
            dropout_p=s.dropout_p,
            stochastic_depth_p=s.stochastic_depth_p,
        )

        block_modules.append(cur_block)

        if _do_add_attention(
            attention_inclusion_cutoff=s.attention_inclusion_cutoff,
            in_features=cur_block.out_features,
            embedding_dim=cur_block.out_feature_sets,
        ):
            cur_attention_block = LCLAttentionBlock(
                embedding_dim=cur_block.out_feature_sets,
                in_features=cur_block.out_features,
            )
            block_modules.append(cur_attention_block)

    logger.debug(
        "No SplitLinear residual blocks specified in CL arguments. Created %d "
        "blocks with final output dimension of %d.",
        len(block_modules),
        cur_size,
    )
    return nn.Sequential(*block_modules)


def _should_break_auto(
    cur_size: int, cutoff: int, direction: Literal["up", "down"]
) -> bool:
    if direction == "down":
        return cur_size <= cutoff
    elif direction == "up":
        return cur_size >= cutoff
    else:
        raise ValueError(f"Unknown direction: {direction}")


class LCLAttentionBlock(nn.Module):
    def __init__(
        self,
        embedding_dim: int,
        in_features: int,
        num_heads: Union[int, Literal["auto"]] = "auto",
        dropout_p: float = 0.0,
        attention_type: Literal["full", "linear"] = "full",
    ):
        super().__init__()

        self.embedding_dim = embedding_dim
        self.in_features = in_features
        self.dropout_p = dropout_p
        self.attention_type = attention_type
        self.out_features = in_features

        if num_heads == "auto":
            self.num_heads = embedding_dim
        else:
            self.num_heads = num_heads
        assert isinstance(self.num_heads, int)

        self.attention: nn.MultiheadAttention | LinearAttention
        if attention_type == "full":
            self.attention = nn.MultiheadAttention(
                embed_dim=self.embedding_dim,
                num_heads=self.num_heads,
                batch_first=True,
                dropout=dropout_p,
            )
        elif attention_type == "linear":
            self.attention = LinearAttention(
                embed_dim=self.embedding_dim,
                heads=self.num_heads,
                dim_head=self.embedding_dim // self.num_heads,
            )

        self.norm = nn.LayerNorm(self.embedding_dim)
        self.pos_emb = PositionalEmbedding(
            embedding_dim=self.embedding_dim,
            max_length=self.in_features // self.embedding_dim,
            dropout=self.dropout_p,
            zero_init=True,
        )

        self.ls = LayerScale(dim=self.in_features, init_values=1e-5)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = x.reshape(x.shape[0], -1, self.embedding_dim)
        out = self.pos_emb(out)

        if self.attention_type == "full":
            out = self.norm(out)
            attn_output, _ = self.attention(out, out, out)
            out = out + attn_output
        elif self.attention_type == "linear":
            out = self.norm(out)
            attn_output = self.attention(out)
            out = out + attn_output
        else:
            raise ValueError("attention_type must be either 'full' or 'linear'")

        out = torch.flatten(out, start_dim=1)
        out = self.ls(out)

        return x + out


def _do_add_attention(
    in_features: int, embedding_dim: int, attention_inclusion_cutoff: Optional[int]
) -> bool:
    if attention_inclusion_cutoff is None:
        return False

    attention_sequence_length = in_features // embedding_dim
    return attention_sequence_length <= attention_inclusion_cutoff



import json
from argparse import Namespace
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Iterable, Literal, Union

import numpy as np
from aislib.misc_utils import ensure_path_exists
from torch.utils.data import DataLoader

from eir import train
from eir.data_load import datasets, label_setup
from eir.data_load.label_setup import al_all_column_ops
from eir.experiment_io.experiment_io import (
    LoadedTrainExperiment,
    check_version,
    get_run_folder_from_model_path,
    load_serialized_train_experiment,
)
from eir.models.model_setup_modules.meta_setup import (
    al_meta_model,
    get_default_meta_class,
    get_meta_model_class_and_kwargs_from_configs,
)
from eir.models.model_setup_modules.model_io import load_model
from eir.models.model_training_utils import get_prediction_outputs_generator
from eir.predict_modules.predict_attributions import compute_predict_attributions
from eir.predict_modules.predict_config import converge_train_and_predict_configs
from eir.predict_modules.predict_data import set_up_default_dataset
from eir.predict_modules.predict_input_setup import set_up_inputs_for_predict
from eir.predict_modules.predict_output_modules.predict_array_output import (
    predict_array_wrapper,
)
from eir.predict_modules.predict_output_modules.predict_sequence_output import (
    predict_sequence_wrapper,
)
from eir.predict_modules.predict_output_modules.predict_tabular_output import (
    predict_tabular_wrapper_no_labels,
    predict_tabular_wrapper_with_labels,
)
from eir.predict_modules.predict_target_setup import (
    MissingTargetsInfo,
    get_target_labels_for_testing,
)
from eir.setup.config import Configs, get_main_parser
from eir.setup.input_setup import al_input_objects_as_dict
from eir.setup.output_setup import al_output_objects_as_dict
from eir.target_setup.target_label_setup import gather_all_ids_from_output_configs
from eir.train import check_dataset_and_batch_size_compatibility
from eir.train_utils.evaluation import (
    deregister_pre_evaluation_hooks,
    register_pre_evaluation_hooks,
    run_all_eval_hook_analysis,
    run_split_evaluation,
)
from eir.train_utils.metrics import al_metric_record_dict, al_step_metric_dict
from eir.train_utils.step_logic import Hooks, prepare_base_batch_default
from eir.train_utils.utils import set_log_level_for_eir_loggers
from eir.utils.logging import get_logger

logger = get_logger(name=__name__, tqdm_compatible=True)

al_predict_hook = Callable[..., dict]
al_predict_hooks = Iterable[al_predict_hook]


@dataclass()
class PredictSpecificCLArgs:
    model_path: str
    evaluate: bool
    output_folder: str
    attribution_background_source: Union[Literal["train"], Literal["predict"]]


def main():
    main_parser = get_main_parser(output_nargs="*", global_nargs="*")

    main_parser.add_argument(
        "--model_path",
        type=str,
        required=True,
        help="Path to model to use for predictions.",
    )

    main_parser.add_argument("--evaluate", dest="evaluate", action="store_true")

    main_parser.add_argument(
        "--output_folder",
        type=str,
        help="Where to save prediction results.",
        required=True,
    )

    main_parser.add_argument(
        "--attribution_background_source",
        type=str,
        help="For attribution analysis, whether to load backgrounds from the data used "
        "for training or to use the current data passed to the predict module.",
        choices=["train", "predict"],
        default="train",
    )

    predict_cl_args = main_parser.parse_args()

    _verify_predict_cl_args(predict_cl_args=predict_cl_args)
    run_folder = get_run_folder_from_model_path(model_path=predict_cl_args.model_path)
    check_version(run_folder=run_folder)

    run_predict(predict_cl_args=predict_cl_args)


def _verify_predict_cl_args(predict_cl_args: Namespace):
    if predict_cl_args.evaluate:
        if len(predict_cl_args.output_configs) == 0:
            raise ValueError(
                "If you want to evaluate, you must specify at least one output config."
                "This is needed to know the target column(s) and values to compute"
                "metrics for (i.e., to compare predicted vs. true values)."
            )


def run_predict(predict_cl_args: Namespace):
    run_folder = get_run_folder_from_model_path(model_path=predict_cl_args.model_path)
    loaded_train_experiment = load_serialized_train_experiment(run_folder=run_folder)

    set_log_level_for_eir_loggers(
        log_level=loaded_train_experiment.configs.gc.vl.log_level
    )

    predict_experiment = get_default_predict_experiment(
        loaded_train_experiment=loaded_train_experiment,
        predict_cl_args=predict_cl_args,
    )

    predict(
        predict_experiment=predict_experiment,
        predict_cl_args=predict_cl_args,
    )

    if predict_experiment.configs.gc.aa.compute_attributions:
        compute_predict_attributions(
            loaded_train_experiment=loaded_train_experiment,
            predict_config=predict_experiment,
        )


def predict(
    predict_experiment: "PredictExperiment",
    predict_cl_args: Namespace,
) -> None:
    hook_finalizers = register_pre_evaluation_hooks(
        model=predict_experiment.model,
        global_config=predict_experiment.configs.global_config,
    )

    output_generator = get_prediction_outputs_generator(
        data_loader=predict_experiment.test_dataloader,
        batch_prep_hook=predict_experiment.hooks.step_func_hooks.base_prepare_batch,
        batch_prep_hook_kwargs={"experiment": predict_experiment},
        model=predict_experiment.model,
        with_labels=predict_cl_args.evaluate,
    )

    criteria = train.get_criteria(outputs_as_dict=predict_experiment.outputs)
    loss_func = train.get_loss_callable(criteria=criteria)

    predict_results = run_split_evaluation(
        output_generator=output_generator,
        output_objects=predict_experiment.outputs,
        experiment_metrics=predict_experiment.metrics,
        loss_function=loss_func,
        device=predict_experiment.configs.gc.be.device,
        with_labels=predict_cl_args.evaluate,
        missing_ids_per_output=predict_experiment.test_dataset.missing_ids_per_output,
    )

    hook_outputs = deregister_pre_evaluation_hooks(
        hook_finalizers=hook_finalizers,
        evaluation_results=predict_results,
    )

    run_all_eval_hook_analysis(
        hook_outputs=hook_outputs,
        run_folder=Path(predict_cl_args.output_folder),
        iteration="predict",
    )

    if predict_cl_args.evaluate:
        serialize_prediction_metrics(
            output_folder=Path(predict_cl_args.output_folder),
            metrics=predict_results.metrics_with_averages,
        )

        predict_tabular_wrapper_with_labels(
            predict_config=predict_experiment,
            all_predictions=predict_results.gathered_outputs,
            all_labels=predict_results.gathered_labels,
            all_ids=predict_results.ids_per_output,
            predict_cl_args=predict_cl_args,
        )
    else:
        predict_tabular_wrapper_no_labels(
            predict_config=predict_experiment,
            all_predictions=predict_results.gathered_outputs,
            all_ids=predict_results.ids_per_output,
            predict_cl_args=predict_cl_args,
        )

    predict_sequence_wrapper(
        predict_experiment=predict_experiment,
        output_folder=predict_cl_args.output_folder,
    )

    predict_array_wrapper(
        predict_experiment=predict_experiment,
        output_folder=predict_cl_args.output_folder,
    )


def serialize_prediction_metrics(output_folder: Path, metrics: al_step_metric_dict):
    ensure_path_exists(path=output_folder, is_folder=True)
    with open(str(output_folder / "calculated_metrics.json"), "w") as outfile:
        parsed = _convert_dict_values_to_python_objects(object_=metrics)
        json.dump(obj=parsed, fp=outfile)


def _convert_dict_values_to_python_objects(object_):
    """
    Needed since json does not serialize numpy dtypes natively.
    """

    if isinstance(object_, np.number):
        return object_.item()
    elif isinstance(object_, dict):
        for key, value in object_.items():
            object_[key] = _convert_dict_values_to_python_objects(object_=value)

    return object_


@dataclass
class PredictExperiment:
    configs: Configs
    inputs: al_input_objects_as_dict
    outputs: al_output_objects_as_dict
    predict_specific_cl_args: PredictSpecificCLArgs
    test_dataset: datasets.DiskDataset | datasets.MemoryDataset
    test_dataloader: DataLoader
    model: al_meta_model
    hooks: "PredictHooks"
    metrics: al_metric_record_dict


@dataclass
class PredictHooks:
    step_func_hooks: "PredictStepFunctionHookStages"
    custom_column_label_parsing_ops: al_all_column_ops = None


@dataclass
class PredictStepFunctionHookStages:
    base_prepare_batch: al_predict_hooks
    model_forward: al_predict_hooks


def get_default_predict_experiment(
    loaded_train_experiment: "LoadedTrainExperiment",
    predict_cl_args: Namespace,
) -> PredictExperiment:
    configs_overloaded_for_predict = converge_train_and_predict_configs(
        train_configs=loaded_train_experiment.configs, predict_cl_args=predict_cl_args
    )

    default_train_hooks = loaded_train_experiment.hooks

    if predict_cl_args.evaluate:
        test_ids = gather_all_ids_from_output_configs(
            output_configs=configs_overloaded_for_predict.output_configs
        )

        label_ops = default_train_hooks.custom_column_label_parsing_ops
        target_labels = get_target_labels_for_testing(
            configs_overloaded_for_predict=configs_overloaded_for_predict,
            custom_column_label_parsing_ops=label_ops,
            ids=test_ids,
        )
        missing_ids_per_output = target_labels.missing_ids_per_output
    else:
        test_ids = label_setup.gather_all_ids_from_all_inputs(
            input_configs=configs_overloaded_for_predict.input_configs
        )
        target_labels = None
        missing_ids_per_output = MissingTargetsInfo(
            missing_ids_per_modality={},
            missing_ids_within_modality={},
            precomputed_missing_ids={},
        )

    test_inputs = set_up_inputs_for_predict(
        test_inputs_configs=configs_overloaded_for_predict.input_configs,
        ids=test_ids,
        hooks=default_train_hooks,
        output_folder=loaded_train_experiment.configs.gc.be.output_folder,
    )

    label_dict = target_labels.label_dict if target_labels else {}
    test_dataset = set_up_default_dataset(
        configs=configs_overloaded_for_predict,
        target_labels_dict=label_dict,
        inputs_as_dict=test_inputs,
        outputs_as_dict=loaded_train_experiment.outputs,
        missing_ids_per_output=missing_ids_per_output,
    )
    predict_batch_size = _auto_set_test_batch_size(
        batch_size=configs_overloaded_for_predict.gc.be.batch_size,
        test_set_size=len(test_dataset),
    )

    check_dataset_and_batch_size_compatibility(
        dataset=test_dataset,
        batch_size=predict_batch_size,
        name="Test",
    )
    test_dataloader = DataLoader(
        dataset=test_dataset,
        batch_size=predict_batch_size,
        shuffle=False,
        num_workers=configs_overloaded_for_predict.gc.be.dataloader_workers,
    )

    func = get_meta_model_class_and_kwargs_from_configs
    fusion_model_class, fusion_model_kwargs = func(
        global_config=configs_overloaded_for_predict.global_config,
        fusion_config=configs_overloaded_for_predict.fusion_config,
        inputs_as_dict=test_inputs,
        outputs_as_dict=loaded_train_experiment.outputs,
        meta_class_getter=get_default_meta_class,
    )

    model = load_model(
        model_path=Path(predict_cl_args.model_path),
        model_class=fusion_model_class,
        model_init_kwargs=fusion_model_kwargs,
        device=configs_overloaded_for_predict.gc.be.device,
        test_mode=True,
        strict_shapes=True,
    )
    assert not model.training

    predict_specific_cl_args = extract_predict_specific_cl_args(
        all_predict_cl_args=predict_cl_args
    )

    default_predict_hooks = _get_default_predict_hooks(train_hooks=default_train_hooks)
    predict_experiment = PredictExperiment(
        configs=configs_overloaded_for_predict,
        inputs=test_inputs,
        outputs=loaded_train_experiment.outputs,
        predict_specific_cl_args=predict_specific_cl_args,
        test_dataset=test_dataset,
        test_dataloader=test_dataloader,
        model=model,
        hooks=default_predict_hooks,
        metrics=loaded_train_experiment.metrics,
    )

    return predict_experiment


def _auto_set_test_batch_size(batch_size: int, test_set_size: int) -> int:
    if test_set_size < batch_size:
        logger.warning(
            f"Test set size ({test_set_size}) is smaller than "
            f"batch size ({batch_size}). "
            f"Setting batch size to test set size."
        )
        batch_size = test_set_size
    return batch_size


def _get_default_predict_hooks(train_hooks: "Hooks") -> PredictHooks:
    stages = PredictStepFunctionHookStages(
        base_prepare_batch=[_hook_default_predict_prepare_batch],
        model_forward=train_hooks.step_func_hooks.model_forward,
    )
    predict_hooks = PredictHooks(
        step_func_hooks=stages,
        custom_column_label_parsing_ops=train_hooks.custom_column_label_parsing_ops,
    )

    return predict_hooks


def _hook_default_predict_prepare_batch(
    experiment: "PredictExperiment",
    loader_batch,
    *args,
    **kwargs,
):
    batch = prepare_base_batch_default(
        loader_batch=loader_batch,
        input_objects=experiment.inputs,
        output_objects=experiment.outputs,
        model=experiment.model,
        device=experiment.configs.gc.be.device,
    )

    state_updates = {"batch": batch}

    return state_updates


def extract_predict_specific_cl_args(
    all_predict_cl_args: Namespace,
) -> PredictSpecificCLArgs:
    predict_module_specific_keys = PredictSpecificCLArgs.__dataclass_fields__.keys()

    dataclass_kwargs = {
        k: v
        for k, v in all_predict_cl_args.__dict__.items()
        if k in predict_module_specific_keys
    }
    predict_specific_cl_args_object = PredictSpecificCLArgs(**dataclass_kwargs)

    return predict_specific_cl_args_object


if __name__ == "__main__":
    main()



from typing import Dict, Literal

import torch
from torch import nn

from eir.models.meta.meta_utils import (
    al_fusion_modules,
    al_input_modules,
    al_output_modules,
    run_meta_forward,
)


class MetaModel(nn.Module):
    def __init__(
        self,
        input_modules: al_input_modules,
        fusion_modules: al_fusion_modules,
        output_modules: al_output_modules,
        fusion_to_output_mapping: Dict[str, Literal["computed", "pass-through"]],
        tensor_broker: nn.ModuleDict,
    ):
        super().__init__()

        self.input_modules = input_modules
        self.fusion_modules = fusion_modules
        self.output_modules = output_modules
        self.fusion_to_output_mapping = fusion_to_output_mapping
        self.tensor_broker = tensor_broker

    def forward(
        self, inputs: dict[str, torch.Tensor]
    ) -> dict[str, dict[str, torch.Tensor]]:

        output_modules_out = run_meta_forward(
            input_modules=self.input_modules,
            fusion_modules=self.fusion_modules,
            output_modules=self.output_modules,
            fusion_to_output_mapping=self.fusion_to_output_mapping,
            inputs=inputs,
        )

        return output_modules_out



from copy import copy
from dataclasses import asdict, dataclass, fields
from pathlib import Path
from types import SimpleNamespace
from typing import TYPE_CHECKING, Any, Dict, Iterable, Optional, Union

import dill
import joblib
from aislib.misc_utils import ensure_path_exists

from eir import __version__
from eir.data_load import label_setup
from eir.data_load.label_setup import (
    al_label_transformers,
    al_label_transformers_object,
)
from eir.setup import schemas
from eir.setup.config import Configs
from eir.setup.output_setup import al_output_objects_as_dict
from eir.train_utils.utils import get_run_folder
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.setup.input_setup import (
        al_input_objects_as_dict,
        al_serializable_input_classes,
        al_serializable_input_objects,
    )
    from eir.train import Experiment
    from eir.train_utils.metrics import al_metric_record_dict
    from eir.train_utils.step_logic import Hooks


logger = get_logger(name=__name__, tqdm_compatible=True)


def get_run_folder_from_model_path(model_path: str) -> Path:
    model_path_object = Path(model_path)
    assert model_path_object.exists()

    run_folder = model_path_object.parents[1]
    assert run_folder.exists()

    return run_folder


@dataclass
class LoadedTrainExperiment:
    configs: Configs
    hooks: "Hooks"
    metrics: "al_metric_record_dict"
    outputs: al_output_objects_as_dict


def load_serialized_train_experiment(run_folder: Path) -> LoadedTrainExperiment:
    check_version(run_folder=run_folder)

    train_config_path = get_train_experiment_serialization_path(run_folder=run_folder)
    with open(train_config_path, "rb") as infile:
        train_config = dill.load(file=infile)

    expected_keys = get_default_experiment_keys_to_serialize()
    train_config_as_dict = train_config.__dict__
    assert set(train_config_as_dict.keys()) == set(expected_keys)

    loaded_experiment = LoadedTrainExperiment(**train_config_as_dict)

    return loaded_experiment


def serialize_experiment(
    experiment: "Experiment",
    run_folder: Path,
    keys_to_serialize: Union[Iterable[str], None],
) -> None:
    serialization_path = get_train_experiment_serialization_path(run_folder=run_folder)
    ensure_path_exists(path=serialization_path)

    filtered_experiment = filter_experiment_by_keys(
        experiment=experiment, keys=keys_to_serialize
    )
    serialize_namespace(namespace=filtered_experiment, output_path=serialization_path)


def get_train_experiment_serialization_path(run_folder: Path) -> Path:
    train_experiment_path = run_folder / "serializations" / "filtered_experiment.dill"

    return train_experiment_path


def filter_experiment_by_keys(
    experiment: "Experiment", keys: Union[None, Iterable[str]] = None
) -> SimpleNamespace:
    filtered = {}

    config_fields = (f.name for f in fields(experiment))
    iterable = keys if keys is not None else config_fields

    for k in iterable:
        filtered[k] = getattr(experiment, k)

    namespace = SimpleNamespace(**filtered)

    return namespace


def serialize_namespace(namespace: SimpleNamespace, output_path: Path) -> None:
    with open(output_path, "wb") as outfile:
        dill.dump(namespace, outfile)


def get_default_experiment_keys_to_serialize() -> Iterable[str]:
    return (
        "configs",
        "outputs",
        "metrics",
        "hooks",
    )


def load_serialized_input_object(
    input_config: schemas.InputConfig,
    input_class: "al_serializable_input_classes",
    *args,
    output_folder: Union[None, str] = None,
    run_folder: Union[None, Path] = None,
    custom_input_name: Union[str, None] = None,
    **kwargs,
) -> "al_serializable_input_objects":
    assert output_folder or run_folder
    if not run_folder:
        assert output_folder is not None
        run_folder = get_run_folder(output_folder=output_folder)

    check_version(run_folder=run_folder)
    input_name = input_config.input_info.input_name
    input_type = input_config.input_info.input_type

    if custom_input_name:
        input_name = custom_input_name

    serialized_input_config_path = get_input_serialization_path(
        run_folder=run_folder,
        input_name=input_name,
        input_type=input_type,
    )

    assert serialized_input_config_path.exists(), serialized_input_config_path
    with open(serialized_input_config_path, "rb") as infile:
        serialized_input_config_object: "al_serializable_input_objects" = dill.load(
            file=infile
        )

    assert isinstance(serialized_input_config_object, input_class)

    train_input_info_kwargs = serialized_input_config_object.__dict__
    assert "input_config" in train_input_info_kwargs.keys()

    loaded_input_info_kwargs = copy(train_input_info_kwargs)

    _check_current_and_loaded_input_config_compatibility(
        current_input_config=input_config,
        loaded_input_config=serialized_input_config_object.input_config,
        serialized_input_config_path=serialized_input_config_path,
    )
    loaded_input_info_kwargs["input_config"] = input_config

    loaded_input_object = input_class(**loaded_input_info_kwargs)

    return loaded_input_object


def get_input_serialization_path(
    run_folder: Path, input_type: str, input_name: str
) -> Path:
    path = (
        run_folder
        / "serializations"
        / f"{input_type}_input_serializations/{input_name}.dill"
    )

    return path


def _check_current_and_loaded_input_config_compatibility(
    current_input_config: schemas.InputConfig,
    loaded_input_config: schemas.InputConfig,
    serialized_input_config_path: Path,
) -> None:
    fieldnames = current_input_config.__dict__.keys()
    assert set(fieldnames) == set(loaded_input_config.__dict__.keys())

    should_be_same = ("model_config",)

    for key in should_be_same:
        current_value = getattr(current_input_config, key)
        loaded_value = getattr(loaded_input_config, key)

        if _should_skip_warning(
            key=key,
            current_value=current_value,
            loaded_value=loaded_value,
        ):
            continue

        if current_value != loaded_value:
            logger.warning(
                "Expected '%s' to be the same in current input configuration"
                "\n'%s'\n and "
                "loaded input configuration \n'%s'\n "
                "(loaded from '%s'). If you are loading"
                " a pretrained EIR module, this can be expected if you are changing "
                "parameters that are expected to be agnostic across runs (e.g. dropout)"
                ", but in many cases this will cause (a) the model you are trying to "
                "load and (b) the model you are setting up for the current experiment "
                "to diverge, which will most likely lead to a RuntimeError. The "
                "resolution is likely to ensure that the input configurations of "
                "(a) and (b) are exactly the same when it comes to model "
                "configurations, which should ensure that the model architectures "
                "match.",
                key,
                current_value,
                loaded_value,
                serialized_input_config_path,
            )


def _should_skip_warning(key: str, current_value: Any, loaded_value: Any) -> bool:
    if key != "model_config":
        return False

    current_dict = asdict(current_value)
    loaded_dict = asdict(loaded_value)

    differing_keys = [k for k, v in current_dict.items() if loaded_dict.get(k) != v]

    return (
        len(differing_keys) == 1
        and differing_keys[0] == "model_type"
        and "linked" in loaded_dict["model_type"]
    )


def serialize_chosen_input_objects(
    inputs_dict: "al_input_objects_as_dict", run_folder: Path
) -> None:
    targets_to_serialize = {"sequence", "bytes", "image", "array"}
    for input_name, input_ in inputs_dict.items():
        input_type = input_.input_config.input_info.input_type

        any_match = any(i for i in targets_to_serialize if input_type == i)

        if any_match:
            output_path = get_input_serialization_path(
                run_folder=run_folder,
                input_type=input_type,
                input_name=input_name,
            )
            ensure_path_exists(path=output_path, is_folder=False)
            with open(output_path, "wb") as outfile:
                dill.dump(obj=input_, file=outfile)


def get_transformer_sources(run_folder: Path) -> Dict[str, list[str]]:
    transformers_to_load = {}
    transformer_sources = run_folder / "serializations/transformers"
    for transformer_source in transformer_sources.iterdir():
        names = sorted([i.stem for i in transformer_source.iterdir()])
        transformers_to_load[transformer_source.stem] = names
    return transformers_to_load


def load_transformer(
    run_folder: Path,
    source_name: str,
    transformer_name: str,
) -> al_label_transformers_object:
    target_transformer_path = label_setup.get_transformer_path(
        run_path=run_folder,
        transformer_name=transformer_name,
        source_name=source_name,
    )
    return joblib.load(filename=target_transformer_path)


def load_transformers(
    transformers_to_load: Optional[Dict[str, list[str]]] = None,
    output_folder: Optional[str] = None,
    run_folder: Optional[Path] = None,
) -> Dict[str, al_label_transformers]:
    if not run_folder and not output_folder:
        raise ValueError("Either 'run_folder' or 'output_folder' must be provided.")

    if not run_folder:
        assert output_folder is not None
        run_folder = get_run_folder(output_folder=output_folder)

    if not transformers_to_load:
        transformers_to_load = get_transformer_sources(run_folder=run_folder)

    loaded_transformers: Dict[str, al_label_transformers] = {}
    for source_name, source_transformers_to_load in transformers_to_load.items():
        loaded_transformers[source_name] = {}
        for transformer_name in source_transformers_to_load:
            loaded_transformers[source_name][transformer_name] = load_transformer(
                run_folder=run_folder,
                source_name=source_name,
                transformer_name=transformer_name,
            )

    return loaded_transformers


def check_version(run_folder: Path) -> None:
    version_file = run_folder / "meta/eir_version.txt"
    if not version_file.exists():
        return

    cur_version = __version__
    with open(version_file, "r") as f:
        loaded_version = f.read().strip()

    if cur_version != loaded_version:
        logger.warning(
            f"The version of EIR used to train this model is {loaded_version}, "
            f"while the current version is {cur_version}. "
            f"This may cause unexpected behavior and subtle bugs."
        )


def get_version_file(run_folder: Path) -> Path:
    return run_folder / "meta/eir_version.txt"



import torch
from torch import Tensor, nn
from torch.nn import functional as F


class LayerScale(nn.Module):
    def __init__(
        self,
        dim: int,
        init_values: float = 1.0,
    ) -> None:
        """
        Note: We use an init value of 1.0 here instead of the generally applied
        1e-5. As we often have linear down sampling layers for residual identities,
        the models might be more inclined to use the main branch, and having the
        scaling too low might cause learning issues.
        """
        super().__init__()
        self.dim = dim
        self.init_values = init_values
        self.gamma = nn.Parameter(
            data=init_values * torch.ones(dim),
            requires_grad=True,
        )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(dim={self.dim}, init_values={self.init_values})"
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x * self.gamma


class GRN(nn.Module):
    """
    GRN (Global Response Normalization) layer
    """

    def __init__(self, in_channels: int):
        super().__init__()

        self.in_channels = in_channels

        self.gamma = nn.Parameter(
            torch.zeros(1, in_channels, 1, 1),
            requires_grad=True,
        )
        self.beta = nn.Parameter(
            torch.zeros(1, in_channels, 1, 1),
            requires_grad=True,
        )

    def __repr__(self):
        return f"{self.__class__.__name__}(in_channels={self.in_channels})"

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        gx = torch.norm(
            input=x,
            p=2.0,
            dim=(2, 3),
            keepdim=True,
        )
        nx = gx / (gx.mean(dim=-1, keepdim=True) + 1e-6)
        return self.gamma * (x * nx) + self.beta + x


class RMSNorm(nn.Module):
    def __init__(
        self,
        dim: int,
        do_scale: bool = True,
        normalize_dim: int = 2,
    ):
        super().__init__()
        self.do_scale = do_scale
        self.normalize_dim = normalize_dim

        self.g: torch.Tensor | int = 1
        if do_scale:
            self.g = nn.Parameter(data=torch.ones(dim), requires_grad=True)

    def forward(self, x: Tensor) -> Tensor:
        normalize_dim = self.normalize_dim

        scale = self.g
        if self.do_scale:
            assert isinstance(self.g, torch.Tensor)
            dims = x.ndim - self.normalize_dim - 1
            scale = append_dims(t=self.g, dims=dims)

        out = (
            F.normalize(input=x, dim=normalize_dim)
            * scale
            * (x.shape[normalize_dim] ** 0.5)
        )
        return out


def append_dims(t: torch.Tensor, dims: int) -> torch.Tensor:
    shape = t.shape
    return t.reshape(*shape, *((1,) * dims))



import logging
import random
from collections import defaultdict
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    DefaultDict,
    Generator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np
import pandas as pd
import seaborn as sns
import torch
from captum._utils.common import (
    _expand_additional_forward_args,
    _expand_target,
    _format_additional_forward_args,
)
from captum._utils.typing import TargetType
from captum.attr import IntegratedGradients
from captum.attr._utils.approximation_methods import approximation_parameters
from captum.attr._utils.common import _reshape_and_sum
from matplotlib import pyplot as plt
from scipy.stats import bootstrap
from torch import Tensor

from eir.setup import schemas
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.data_load.label_setup import al_label_transformers_object
    from eir.interpretation.interpretation import SampleAttribution


logger = get_logger(name=__name__)


def get_target_class_name(
    sample_label: torch.Tensor,
    target_transformer: "al_label_transformers_object",
    column_type: str,
    target_column_name: str,
):
    if column_type == "con":
        return target_column_name

    tt_it = target_transformer.inverse_transform
    cur_trn_label = tt_it([sample_label.item()])[0]

    return cur_trn_label


def stratify_attributions_by_target_classes(
    all_attributions: Sequence["SampleAttribution"],
    target_transformer: "al_label_transformers_object",
    output_name: str,
    target_column: str,
    column_type: str,
) -> DefaultDict[Any, list["SampleAttribution"]]:
    all_attributions_target_class_stratified = defaultdict(list)

    for sample in all_attributions:
        cur_labels_all = sample.sample_info.target_labels
        cur_labels = cur_labels_all[output_name][target_column]
        cur_label_name = get_target_class_name(
            sample_label=cur_labels,
            target_transformer=target_transformer,
            column_type=column_type,
            target_column_name=target_column,
        )
        all_attributions_target_class_stratified[cur_label_name].append(sample)

    return all_attributions_target_class_stratified


def plot_attributions_bar(
    df_attributions: pd.DataFrame,
    output_path: Union[str, Path],
    top_n: int = 20,
    title: str = "",
    use_bootstrap: bool = True,
) -> None:
    df_token_top_n = calculate_top_n_tokens(
        df_attributions=df_attributions,
        top_n=top_n,
        use_bootstrap=use_bootstrap,
    )
    df_attributions_top_n_sorted = df_attributions[
        df_attributions["Input"].isin(df_token_top_n.index)
    ]

    order = (
        df_attributions_top_n_sorted.groupby("Input")
        .mean()
        .sort_values(by="Attribution", ascending=False)
        .index
    )

    ax: plt.Axes = sns.barplot(
        data=df_attributions_top_n_sorted,
        x="Attribution",
        y="Input",
        order=order,
        color="#1f77b4",
        capsize=0.1,
        errorbar=("ci", 95),
    )

    plt.tight_layout()
    sns_figure: Optional[plt.Figure] = ax.get_figure()
    assert sns_figure is not None

    if title:
        ax.set_title(title)

    sns_figure.set_size_inches(10, 0.5 * top_n)
    sns_figure.savefig(fname=output_path, bbox_inches="tight")

    plt.close("all")


def calculate_token_statistics(
    df_attributions: pd.DataFrame,
    use_bootstrap: bool = True,
    n_bootstraps: int = 1000,
) -> pd.DataFrame:
    df_token_stats = pd.DataFrame()
    log_once_callable = get_log_once_callable(logger_=logger)

    for input_feature in df_attributions["Input"].unique():
        feature_attributions = df_attributions[
            df_attributions["Input"] == input_feature
        ]["Attribution"].to_numpy()

        if use_bootstrap:

            if len(np.unique(feature_attributions)) < 2:
                log_once_callable(
                    f"Feature '{input_feature}' has less than 2 unique attribution "
                    f"values; unable to perform bootstrapping. Calculating mean "
                    f"without bootstrapping and setting CI_Size to 0 as a default. "
                    f"There might be other input features with the same issue, this "
                    f"message will only be logged once per attribution calculation."
                )

                df_token_stats.loc[input_feature, "AttributionMean"] = np.mean(
                    feature_attributions
                )
                df_token_stats.loc[input_feature, "CI_Size"] = np.nan
                continue

            res = bootstrap(
                (feature_attributions,),
                np.mean,
                n_resamples=n_bootstraps,
                method="basic",
            )
            mean = np.mean(feature_attributions)
            ci_low, ci_high = res.confidence_interval.low, res.confidence_interval.high
            ci_size = ci_high - ci_low

            df_token_stats.loc[input_feature, "AttributionMean"] = mean
            df_token_stats.loc[input_feature, "CI_Size"] = ci_size
        else:
            mean = feature_attributions.mean()
            df_token_stats.loc[input_feature, "AttributionMean"] = mean
            df_token_stats.loc[input_feature, "CI_Size"] = 0

    return df_token_stats


def get_log_once_callable(logger_: logging.Logger) -> Callable:
    has_logged = False

    def _do_log_once(msg: str):
        nonlocal has_logged
        if not has_logged:
            logger_.info(msg)
            has_logged = True

    return _do_log_once


def calculate_top_n_tokens(
    df_attributions: pd.DataFrame,
    top_n: int,
    use_bootstrap: bool,
) -> pd.DataFrame:
    df_token_stats = calculate_token_statistics(
        df_attributions=df_attributions,
        use_bootstrap=use_bootstrap,
    )

    if use_bootstrap:
        df_token_stats["AttributionScore"] = df_token_stats["AttributionMean"] - (
            0.5 * df_token_stats["CI_Size"]
        )
    else:
        df_token_stats["AttributionScore"] = df_token_stats["AttributionMean"]

    df_token_stats_top_n = df_token_stats.nlargest(top_n, "AttributionScore")
    df_token_stats_top_n_sorted = df_token_stats_top_n.sort_values(
        by="AttributionScore", ascending=False
    )

    return df_token_stats_top_n_sorted


def filter_and_sort_attributions(
    df_attributions: pd.DataFrame, df_token_top_n: pd.DataFrame
) -> pd.DataFrame:
    df_attributions_top_n = df_attributions[
        df_attributions["Input"].isin(df_token_top_n.index)
    ]

    df_attributions_top_n_sorted = df_attributions_top_n.sort_values(
        by="Attribution", ascending=False
    )

    return df_attributions_top_n_sorted


def get_basic_sample_attributions_to_analyse_generator(
    interpretation_config: schemas.BasicInterpretationConfig,
    all_attributions: list["SampleAttribution"],
) -> Generator["SampleAttribution", None, None]:
    strategy = interpretation_config.interpretation_sampling_strategy
    n_samples = interpretation_config.num_samples_to_interpret

    if strategy == "first_n":
        base = all_attributions[:n_samples]
    elif strategy == "random_sample":
        base = random.sample(all_attributions, n_samples)
    else:
        raise ValueError(f"Unrecognized option for sequence sampling: {strategy}.")

    manual_samples = interpretation_config.manual_samples_to_interpret
    if manual_samples:
        for attribution in all_attributions:
            if attribution.sample_info.ids in manual_samples:
                base.append(attribution)

    for item in base:
        yield item


class MyIntegratedGradients(IntegratedGradients):
    """
    Only change here is casting the input to float32 for MPS devices.
    """

    def _attribute(
        self,
        inputs: Tuple[Tensor, ...],
        baselines: Tuple[Union[Tensor, int, float], ...],
        target: TargetType = None,
        additional_forward_args: Any = None,
        n_steps: int = 50,
        method: str = "gausslegendre",
        step_sizes_and_alphas: Union[None, Tuple[List[float], List[float]]] = None,
    ) -> Tuple[Tensor, ...]:
        if step_sizes_and_alphas is None:
            # retrieve step size and scaling factor for specified
            # approximation method
            step_sizes_func, alphas_func = approximation_parameters(method)
            step_sizes, alphas = step_sizes_func(n_steps), alphas_func(n_steps)
        else:
            step_sizes, alphas = step_sizes_and_alphas

        # scale features and compute gradients. (batch size is abbreviated as bsz)
        # scaled_features' dim -> (bsz * #steps x inputs[0].shape[1:], ...)
        scaled_features_tpl = tuple(
            torch.cat(
                [baseline + alpha * (input - baseline) for alpha in alphas], dim=0
            ).requires_grad_()
            for input, baseline in zip(inputs, baselines)
        )

        additional_forward_args = _format_additional_forward_args(
            additional_forward_args
        )
        # apply number of steps to additional forward args
        # currently, number of steps is applied only to additional forward arguments
        # that are nd-tensors. It is assumed that the first dimension is
        # the number of batches.
        # dim -> (bsz * #steps x additional_forward_args[0].shape[1:], ...)
        input_additional_args = (
            _expand_additional_forward_args(additional_forward_args, n_steps)
            if additional_forward_args is not None
            else None
        )
        expanded_target = _expand_target(target, n_steps)

        # grads: dim -> (bsz * #steps x inputs[0].shape[1:], ...)
        grads = self.gradient_func(
            forward_fn=self.forward_func,
            inputs=scaled_features_tpl,
            target_ind=expanded_target,
            additional_forward_args=input_additional_args,
        )

        # flattening grads so that we can multiply it with step-size
        # calling contiguous to avoid `memory whole` problems
        scaled_grads = [
            grad.contiguous().view(n_steps, -1)
            * torch.tensor(step_sizes, dtype=torch.float32)  # here cast for MPS
            .view(n_steps, 1)
            .to(grad.device)
            for grad in grads
        ]

        # aggregates across all steps for each tensor in the input tuple
        # total_grads has the same dimensionality as inputs
        total_grads = tuple(
            _reshape_and_sum(
                scaled_grad, n_steps, grad.shape[0] // n_steps, grad.shape[1:]
            )
            for (scaled_grad, grad) in zip(scaled_grads, grads)
        )

        # computes attribution for each tensor in input tuple
        # attributions has the same dimensionality as inputs
        if not self.multiplies_by_inputs:
            attributions = total_grads
        else:
            attributions = tuple(
                total_grad * (input - baseline)
                for total_grad, input, baseline in zip(total_grads, inputs, baselines)
            )
        return attributions


def get_long_format_attribution_df(
    parsed_attributions: dict[str, list[float]]
) -> pd.DataFrame:
    if not isinstance(parsed_attributions, dict):
        raise TypeError("Input must be a dictionary")

    if not parsed_attributions:
        return pd.DataFrame(columns=["Input", "Attribution"])

    series_dict: dict[str, pd.Series] = {}
    for k, v in parsed_attributions.items():
        if not isinstance(v, list):
            raise ValueError(f"Value for key '{k}' must be a list")
        if not all(isinstance(x, (int, float, np.float32, np.float64)) for x in v):
            raise ValueError(f"All values in list {v} for key '{k}' must be numbers")
        series_dict[k] = pd.Series(v)

    df: pd.DataFrame = pd.concat(series_dict)
    df = df.reset_index(level=0).reset_index(drop=True)

    df = df.rename(columns={df.columns[0]: "Input", df.columns[1]: "Attribution"})

    return df



from typing import Dict, List, Optional, Sequence


class Vocab:
    def __init__(self, tokens: Optional[List[str]] = None):
        self.itos: list[str] = []  # index to string
        self.stoi: dict[str, int] = {}  # string to index
        self.default_index: int = -1

        if tokens:
            for token in tokens:
                self.append_token(token)

    def __len__(self) -> int:
        return len(self.itos)

    def __contains__(self, token: str) -> bool:
        return token in self.stoi

    def __getitem__(self, token: str) -> int:
        return self.stoi.get(token, self.default_index)

    def __call__(self, tokens: Sequence[str]) -> List[int]:
        """
        Returns the indices associated with a list of tokens.

        Args:
            tokens: A list of tokens to lookup.

        Returns:
            A list of indices corresponding to the input tokens.
        """
        return self.lookup_indices(list(tokens))

    def set_default_index(self, index: int) -> None:
        self.default_index = index

    def get_default_index(self) -> Optional[int]:
        return self.default_index

    def insert_token(self, token: str, index: int) -> None:
        if index < 0 or index > len(self.itos):
            raise ValueError(f"Index {index} out of range [0, {len(self.itos)}]")
        if token in self.stoi:
            raise RuntimeError(f"Token '{token}' already exists in the vocab")

        self.itos.insert(index, token)
        self.stoi[token] = index
        for i in range(index + 1, len(self.itos)):
            self.stoi[self.itos[i]] = i

    def append_token(self, token: str) -> None:
        if token in self.stoi:
            raise RuntimeError(f"Token '{token}' already exists in the vocab")
        self.itos.append(token)
        self.stoi[token] = len(self.itos) - 1

    def lookup_token(self, index: int) -> str:
        if index < 0 or index >= len(self.itos):
            raise IndexError(f"Index {index} out of range [0, {len(self.itos)})")
        return self.itos[index]

    def lookup_tokens(self, indices: List[int]) -> List[str]:
        return [self.lookup_token(index) for index in indices]

    def lookup_indices(self, tokens: List[str]) -> List[int]:
        return [self[token] for token in tokens]

    def get_stoi(self) -> Dict[str, int]:
        return self.stoi.copy()

    def get_itos(self) -> List[str]:
        return self.itos.copy()



import base64
from io import BytesIO
from typing import TYPE_CHECKING, Any, Dict, Literal, Optional, Sequence, Union

import numpy as np
import numpy.typing as npt
import torch
from PIL import Image
from sklearn.preprocessing import StandardScaler

from eir.data_load.data_streaming.streaming_dataset_utils import (
    streamline_sequence_manual_data,
)
from eir.data_load.label_setup import (
    al_label_transformers,
    al_label_transformers_object,
)
from eir.predict_modules.predict_tabular_input_setup import (
    ComputedPredictTabularInputInfo,
)
from eir.serve_modules.serve_schemas import ComputedServeTabularInputInfo
from eir.setup.input_setup_modules.setup_array import ComputedArrayInputInfo
from eir.setup.input_setup_modules.setup_bytes import ComputedBytesInputInfo
from eir.setup.input_setup_modules.setup_image import ComputedImageInputInfo
from eir.setup.input_setup_modules.setup_omics import ComputedOmicsInputInfo
from eir.setup.input_setup_modules.setup_sequence import ComputedSequenceInputInfo
from eir.setup.input_setup_modules.setup_tabular import ComputedTabularInputInfo
from eir.setup.schemas import ImageInputDataConfig, SequenceInputDataConfig
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.setup.input_setup import al_input_objects_as_dict


logger = get_logger(name=__name__, tqdm_compatible=True)

al_inputs_prepared = dict[
    str,
    np.ndarray | torch.Tensor | list[str] | str | dict | Image.Image,
]


def prepare_request_input_data_wrapper(
    request_data: Sequence[dict[str, Any]],
    input_objects: "al_input_objects_as_dict",
) -> Sequence[al_inputs_prepared]:
    all_inputs_prepared = []

    for request in request_data:
        inputs_prepared = prepare_request_input_data(
            request_data=request,
            input_objects=input_objects,
        )
        all_inputs_prepared.append(inputs_prepared)

    return all_inputs_prepared


def prepare_request_input_data(
    request_data: Dict[str, Any],
    input_objects: "al_input_objects_as_dict",
) -> Dict[str, Any]:
    inputs_prepared: al_inputs_prepared = {}

    for name, serialized_data in request_data.items():
        input_object = input_objects[name]
        input_type = input_object.input_config.input_info.input_type
        input_type_info = input_object.input_config.input_type_info

        match input_object:
            case ComputedOmicsInputInfo():
                assert input_type == "omics"
                shape = input_object.data_dimensions.full_shape()[1:]
                array_np = deserialize_array(
                    array_str=serialized_data,
                    dtype=np.bool_,
                    shape=shape,
                )
                assert len(array_np.shape) == 2
                array_raw = torch.from_numpy(array_np)

                inputs_prepared[name] = array_raw

            case ComputedSequenceInputInfo():
                assert input_type == "sequence"
                assert isinstance(input_type_info, SequenceInputDataConfig)

                sequence_streamlined = streamline_sequence_manual_data(
                    data=serialized_data,
                    split_on=input_type_info.split_on,
                )

                inputs_prepared[name] = sequence_streamlined

            case ComputedBytesInputInfo():
                assert input_type == "bytes"
                array_np = deserialize_array(
                    array_str=serialized_data,
                    dtype=np.uint8,
                    shape=(-1,),
                )
                array_raw = torch.from_numpy(array_np).to(dtype=torch.int64)
                inputs_prepared[name] = array_raw

            case ComputedImageInputInfo():
                assert input_type == "image"
                assert isinstance(input_type_info, ImageInputDataConfig)
                image_data = deserialize_image(
                    image_str=serialized_data,
                    image_mode=input_type_info.mode,
                )
                inputs_prepared[name] = image_data

            case (
                ComputedTabularInputInfo()
                | ComputedPredictTabularInputInfo()
                | ComputedServeTabularInputInfo()
            ):
                assert input_type == "tabular"
                transformers = input_object.labels.label_transformers
                tabular_data = _streamline_tabular_request_data(
                    tabular_input=serialized_data, transformers=transformers
                )
                inputs_prepared[name] = tabular_data

            case ComputedArrayInputInfo():
                assert input_type == "array"
                array_np = deserialize_array(
                    array_str=serialized_data,
                    dtype=input_object.dtype,
                    shape=input_object.data_dimensions.full_shape(),
                )
                inputs_prepared[name] = array_np

            case _:
                raise ValueError(f"Unknown input type '{input_type}'")

    return inputs_prepared


def _streamline_tabular_request_data(
    tabular_input: Dict, transformers: al_label_transformers
) -> Dict:
    parsed_output = {}
    for name, value in tabular_input.items():
        cur_transformer = transformers[name]

        value_transformed = _parse_transformer_output(
            transformer=cur_transformer,
            value=value,
        )

        parsed_output[name] = value_transformed

    return parsed_output


def _parse_transformer_output(
    transformer: al_label_transformers_object, value: float | int
) -> Union[float, int]:
    value_parsed: list
    if isinstance(transformer, StandardScaler):
        value_parsed = [[value]]
    else:
        value_parsed = [value]

    value_transformed = transformer.transform(value_parsed)
    if isinstance(transformer, StandardScaler):
        value_transformed = value_transformed[0][0]
    else:
        value_transformed = value_transformed[0]

    assert not isinstance(value_transformed, list)

    return value_transformed


def deserialize_array(array_str: str, dtype: npt.DTypeLike, shape: tuple) -> np.ndarray:
    array_bytes = base64.b64decode(array_str)
    return np.frombuffer(array_bytes, dtype=dtype).reshape(shape).copy()


def deserialize_image(
    image_str: str, image_mode: Optional[Literal["L", "RGB", "RGBA"]]
) -> Image.Image:
    """
    Note we convert to RGB to be compatible with the default_loader.
    """
    image_data = base64.b64decode(image_str)
    image = Image.open(BytesIO(image_data))

    if image_mode is not None:
        return image.convert(image_mode)

    return image



from typing import TYPE_CHECKING, Dict, Literal, MutableMapping, NewType, Protocol

import torch

if TYPE_CHECKING:
    from eir.models.fusion.fusion import al_fused_features, al_fusion_model_configs

FeatureExtractorOutType = NewType("FeatureExtractorOutType", torch.Tensor)


class FeatureExtractorProtocol(Protocol):
    @property
    def num_out_features(self) -> int: ...

    def __call__(self, input: torch.Tensor) -> FeatureExtractorOutType: ...


class FeatureExtractorProtocolWithL1(Protocol):
    @property
    def num_out_features(self) -> int: ...

    @property
    def l1_penalized_weights(self) -> torch.Tensor: ...

    def __call__(self, input: torch.Tensor) -> FeatureExtractorOutType: ...


class FusionModuleProtocol(Protocol):
    model_config: "al_fusion_model_configs"

    @property
    def fusion_in_dim(self) -> int: ...

    @property
    def num_out_features(self) -> int: ...

    def __call__(
        self, input: Dict[str, FeatureExtractorOutType]
    ) -> "al_fused_features": ...


class OutputModuleProtocol(Protocol):
    def __call__(
        self,
        input: "al_fused_features",
    ) -> Dict[str, torch.Tensor]: ...


al_input_modules = MutableMapping[
    str, FeatureExtractorProtocolWithL1 | FeatureExtractorProtocol
]
al_fusion_modules = MutableMapping[str, FusionModuleProtocol]
al_output_modules = MutableMapping[str, OutputModuleProtocol]


def run_meta_forward(
    input_modules: al_input_modules,
    fusion_modules: al_fusion_modules,
    output_modules: al_output_modules,
    fusion_to_output_mapping: Dict[str, Literal["computed", "pass-through"]],
    inputs: dict[str, torch.Tensor],
) -> dict[str, dict[str, torch.Tensor]]:

    feature_extractors_out = {}
    for module_name, cur_input_module in input_modules.items():
        module_input = inputs[module_name]
        feature_extractors_out[module_name] = cur_input_module(module_input)

    fused_features = {}
    for output_type, fusion_module in fusion_modules.items():
        fused_features[output_type] = fusion_module(feature_extractors_out)

    output_modules_out = {}
    for output_name, output_module in output_modules.items():
        cur_fusion_target = fusion_to_output_mapping[output_name]
        corresponding_fused_features = fused_features[cur_fusion_target]

        key = f"__extras_{output_name}"
        if key in inputs:
            corresponding_fused_features[key] = inputs[key]

        cur_output = output_module(corresponding_fused_features)
        output_modules_out[output_name] = cur_output

    return output_modules_out



import pickle
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Protocol, Tuple

import numpy as np
import pandas as pd
from aislib.misc_utils import ensure_path_exists

from eir.interpretation.interpretation_utils import get_target_class_name
from eir.setup.input_setup_modules.setup_omics import (
    ComputedOmicsInputInfo,
    read_bim,
    read_subset_file,
)
from eir.setup.schemas import OmicsInputDataConfig
from eir.utils.logging import get_logger
from eir.visualization import interpretation_visualization as av

if TYPE_CHECKING:
    from eir.data_load.label_setup import al_label_transformers_object
    from eir.interpretation.interpretation import SampleAttribution
    from eir.train import Experiment

al_gradients_dict = Dict[str, List[np.ndarray]]
al_top_gradients_dict = Dict[str, Dict[str, np.ndarray]]
al_scaled_grads_dict = Dict[str, Dict[str, np.ndarray]]
al_transform_funcs = Dict[str, Tuple[Callable]]

logger = get_logger(name=__name__, tqdm_compatible=True)


@dataclass
class ParsedOmicsAttributions:
    accumulated_acts: Dict[str, np.ndarray]
    accumulated_acts_masked: Dict[str, np.ndarray]


class OmicsConsumerCallable(Protocol):
    def __call__(
        self,
        attribution: Optional["SampleAttribution"],
    ) -> Optional[ParsedOmicsAttributions]: ...


def get_omics_consumer(
    target_transformer: "al_label_transformers_object",
    input_name: str,
    output_name: str,
    target_column: str,
    column_type: str,
) -> OmicsConsumerCallable:
    acc_acts: dict[str, np.ndarray] = {}
    acc_acts_masked: dict[str, np.ndarray] = {}
    n_samples: dict[str, int] = {}

    def _consumer(
        attribution: Optional["SampleAttribution"],
    ) -> Optional[ParsedOmicsAttributions]:
        nonlocal n_samples

        if attribution is None:
            for key, value in acc_acts.items():
                acc_acts[key] = value / n_samples[key]

            for key, value in acc_acts_masked.items():
                acc_acts_masked[key] = value / n_samples[key]

            parsed_attributions = ParsedOmicsAttributions(
                accumulated_acts=acc_acts, accumulated_acts_masked=acc_acts_masked
            )
            return parsed_attributions

        sample_inputs = attribution.sample_info.inputs
        sample_target_labels = attribution.sample_info.target_labels

        cur_label_name = get_target_class_name(
            sample_label=sample_target_labels[output_name][target_column],
            target_transformer=target_transformer,
            column_type=column_type,
            target_column_name=target_column,
        )

        sample_acts = attribution.sample_attributions[input_name].squeeze()
        if cur_label_name not in acc_acts:
            acc_acts[cur_label_name] = sample_acts
        else:
            acc_acts[cur_label_name] += sample_acts

        # we want to keep the original sample for masking
        inputs_omics = sample_inputs[input_name]
        single_sample_copy = deepcopy(inputs_omics).cpu().numpy()
        single_acts_masked = (single_sample_copy * sample_acts).squeeze()

        if cur_label_name not in acc_acts_masked:
            acc_acts_masked[cur_label_name] = single_acts_masked
            n_samples[cur_label_name] = 1
        else:
            acc_acts_masked[cur_label_name] += single_acts_masked
            n_samples[cur_label_name] += 1

        return None

    return _consumer


def analyze_omics_input_attributions(
    experiment: "Experiment",
    input_name: str,
    target_column_name: str,
    target_column_type: str,
    attribution_outfolder: Path,
    all_attributions: ParsedOmicsAttributions,
) -> None:
    exp = experiment

    acc_acts = all_attributions.accumulated_acts
    acc_acts_masked = all_attributions.accumulated_acts_masked

    abs_grads = True if target_column_type == "con" else False
    top_gradients_dict = get_snp_cols_w_top_grads(
        accumulated_grads=acc_acts, abs_grads=abs_grads
    )

    input_object = exp.inputs[input_name]
    assert isinstance(input_object, ComputedOmicsInputInfo)

    omics_data_type_config = input_object.input_config.input_type_info
    assert isinstance(omics_data_type_config, OmicsInputDataConfig)

    assert omics_data_type_config.snp_file is not None
    cur_snp_path = Path(omics_data_type_config.snp_file)
    df_snps = read_bim(bim_file_path=str(cur_snp_path))

    if omics_data_type_config.subset_snps_file:
        subset_snps = read_subset_file(
            subset_snp_file_path=omics_data_type_config.subset_snps_file
        )
        df_snps = df_snps[df_snps["VAR_ID"].isin(subset_snps)]

    classes = sorted(list(top_gradients_dict.keys()))
    scaled_grads = gather_and_rescale_snps(
        all_gradients_dict=acc_acts,
        top_gradients_dict=top_gradients_dict,
        classes=classes,
    )
    av.plot_top_gradients(
        gathered_scaled_grads=scaled_grads,
        top_gradients_dict=top_gradients_dict,
        df_snps=df_snps,
        output_folder=attribution_outfolder,
    )

    _save_omics_grads(
        sample_output_folder=attribution_outfolder,
        file_name="top_acts.npy",
        grads_dict=top_gradients_dict,
    )

    save_masked_grads(
        acc_grads_times_inp=acc_acts_masked,
        top_gradients_dict=top_gradients_dict,
        df_snps=df_snps,
        sample_outfolder=attribution_outfolder,
    )

    df_snp_grads = _save_snp_gradients(
        accumulated_grads=acc_acts, outfolder=attribution_outfolder, df_snps=df_snps
    )
    df_snp_grads_with_abs = _add_absolute_summed_snp_gradients_to_df(
        df_snp_grads=df_snp_grads
    )
    av.plot_snp_manhattan_plots(
        df_snp_grads=df_snp_grads_with_abs,
        outfolder=attribution_outfolder,
        title_extra=f" - {target_column_name}",
    )


def _save_snp_gradients(
    accumulated_grads: Dict[str, np.ndarray], outfolder: Path, df_snps: pd.DataFrame
) -> pd.DataFrame:
    df_output = deepcopy(df_snps)
    for label, grads in accumulated_grads.items():
        grads_np = grads
        grads_averaged = grads_np.sum(0)

        column_name = label + "_attributions"
        df_output[column_name] = grads_averaged

    df_output.to_csv(path_or_buf=outfolder / "snp_attributions.csv")
    return df_output


def _add_absolute_summed_snp_gradients_to_df(
    df_snp_grads: pd.DataFrame,
) -> pd.DataFrame:
    df_snp_grads_copy = df_snp_grads.copy()

    attributions_columns = [
        i for i in df_snp_grads_copy.columns if i.endswith("_attributions")
    ]
    if "Aggregated_attributions" in attributions_columns:
        raise ValueError(
            "Cannot compute aggregated attributions as reserved column name "
            "'Aggregated' already present as target (all attribution columns: %s). "
            "Please rename the column 'Aggregated' to something else "
            "in the relevant target file.",
            attributions_columns,
        )

    df_snp_grads_copy["Aggregated_attributions"] = (
        df_snp_grads_copy[attributions_columns].abs().sum(axis=1)
    )

    return df_snp_grads_copy


def rescale_gradients(gradients: np.ndarray) -> np.ndarray:
    gradients_resc = gradients - gradients.min()
    gradients_resc = gradients_resc / (gradients.max() - gradients.min())
    return gradients_resc


def get_snp_cols_w_top_grads(
    accumulated_grads: Dict[str, np.ndarray],
    n: int = 10,
    custom_indexes_dict: Optional[dict[str, np.ndarray]] = None,
    abs_grads: bool = False,
) -> Dict[str, Dict[str, np.ndarray]]:
    """
    `accumulated_grads` specs:

        {'Asia': [all grads for samples w. Asia label], ...}

    First we find the average gradients for each SNP form by averaging accross
    samples.

    Then we have something that is (4 x width) dimension, and we find the per
    column (i.e. SNP form) max gradient. Of the max grad per SNP, we want to
    find the top N SNPs where the maximum gradient (any form) is found.

    We use those indexes to grab the top SNPs and grads per class.
    """
    top_snps_per_class: Dict[str, Dict[str, np.ndarray]] = {}

    for cls, grads in accumulated_grads.items():
        if grads is not None:
            top_snps_per_class[cls] = {}

            if not custom_indexes_dict:
                if abs_grads:
                    grads = np.abs(grads)

                sum_snp_values = grads.sum(0)

                top_n_idxs = sorted(np.argpartition(sum_snp_values, -n)[-n:])
                top_n_idxs_np: np.ndarray = np.array(top_n_idxs)

                top_snps_per_class[cls]["top_n_idxs"] = top_n_idxs_np
                top_snps_per_class[cls]["top_n_grads"] = grads[:, top_n_idxs_np]

            else:
                top_n_idxs_np = custom_indexes_dict[cls]
                top_snps_per_class[cls]["top_n_idxs"] = top_n_idxs_np
                top_snps_per_class[cls]["top_n_grads"] = grads[:, top_n_idxs_np]
        else:
            logger.warning(
                "No gradients aggregated for class %s due to no "
                "correct predictions for the class, top attributions "
                "will not be plotted.",
                cls,
            )

    return top_snps_per_class


def gather_and_rescale_snps(
    all_gradients_dict: Dict[str, np.ndarray],
    top_gradients_dict: al_top_gradients_dict,
    classes: List[str],
) -> al_scaled_grads_dict:
    """
    `accumulated_grads` specs:

        {'Asia': [all grads for samples w. Asia label],
         'Europe': [all grads for samples w. Europe label]}

    `top_gradients_dict` specs:

        {'Asia': {'top_n_idxs': [],
                  'top_n_grads': [],
                 }
        }

    We want to create a dict with the following specs:

        {
            'Asia': {
                        'Asia Grads Indexed By Top Asia SNPs': [...],
                        'Europe Grads Indexed By Top Asia SNPs': [...],
                        'Africa Grads Indexed By Top Asia SNPs': [...],
                    }
            'Europe': {
                        'Asia Grads Indexed By Top Europe SNPs': [...],
                        'Europe Grads Indexed By Top Europe SNPs': [...],
                        'Africa Grads Indexed By Top Europe SNPs': [...],
                    }
        }

    The each row of the above is a label (i.e. country, e.g. 'Asia') and each
    column contains gradient values as indexed by the top gradient SNPs for
    that label (i.e. highest positive gradients sum over Asian samples).

    We make sure that the top gradients gotten in this function are the
    same as gotten previously (when we just look at each country separately).

    We then accumulate each column (average grads for each country, indexed
    by the top grads for a given country), and rescale them (so we have
    column-wise rescaling).
    """
    top_snps_dict: al_scaled_grads_dict = {}

    for col_name_ in classes:
        cls_top_idxs_ = top_gradients_dict[col_name_]["top_n_idxs"]
        top_snps_dict[col_name_] = {}

        cur_top_snps_all_labels = []
        for row_name_ in classes:
            cur_row_grads_arr = all_gradients_dict[row_name_]
            cur_row_grads_mean = cur_row_grads_arr
            cur_row_indexed_grads = cur_row_grads_mean[:, cls_top_idxs_]

            cur_top_snps_all_labels.append(cur_row_indexed_grads)

        cur_top_snps_all_labels_np = np.array(cur_top_snps_all_labels)
        cur_top_snps_all_labels_rescaled = rescale_gradients(
            gradients=cur_top_snps_all_labels_np
        )

        for list_idx, row_name_ in enumerate(classes):
            cur_rescaled = cur_top_snps_all_labels_rescaled[list_idx]
            top_snps_dict[col_name_][row_name_] = cur_rescaled

    return top_snps_dict


def save_masked_grads(
    acc_grads_times_inp: Dict[str, np.ndarray],
    top_gradients_dict: al_top_gradients_dict,
    df_snps: pd.DataFrame,
    sample_outfolder: Path,
) -> None:
    top_grads_msk_inputs = index_masked_grads(
        top_grads_dict=top_gradients_dict,
        accumulated_grads_times_input=acc_grads_times_inp,
    )

    classes = sorted(list(top_gradients_dict.keys()))
    scaled_grads = gather_and_rescale_snps(
        all_gradients_dict=acc_grads_times_inp,
        top_gradients_dict=top_grads_msk_inputs,
        classes=classes,
    )
    av.plot_top_gradients(
        gathered_scaled_grads=scaled_grads,
        top_gradients_dict=top_grads_msk_inputs,
        df_snps=df_snps,
        output_folder=sample_outfolder,
        fname="top_snps_masked.pdf",
    )

    _save_omics_grads(
        sample_output_folder=sample_outfolder,
        file_name="top_acts_masked.npy",
        grads_dict=top_grads_msk_inputs,
    )


def index_masked_grads(
    top_grads_dict: al_top_gradients_dict,
    accumulated_grads_times_input: Dict[str, np.ndarray],
) -> al_top_gradients_dict:
    indexes_from_all_grads = {
        key: top_grads_dict[key]["top_n_idxs"] for key in top_grads_dict.keys()
    }

    top_gradients_dict_masked_inputs = get_snp_cols_w_top_grads(
        accumulated_grads_times_input, custom_indexes_dict=indexes_from_all_grads
    )

    return top_gradients_dict_masked_inputs


def _save_omics_grads(
    sample_output_folder: Path,
    file_name: str,
    grads_dict: dict[str, dict[str, np.ndarray]],
) -> None:
    output_path = sample_output_folder / file_name
    ensure_path_exists(path=output_path)

    with open(output_path, "wb") as f:
        pickle.dump(obj=grads_dict, file=f)



from typing import TYPE_CHECKING, Dict, Type

import torch

from eir.models.output.sequence.sequence_output_modules import SequenceOutputModule
from eir.setup.output_setup_modules.sequence_output_setup import (
    ComputedSequenceOutputInfo,
)

al_sequence_module_classes = Type[SequenceOutputModule]

if TYPE_CHECKING:
    from eir.models.model_setup_modules.meta_setup import FeatureExtractorInfo


def get_sequence_output_module_from_model_config(
    output_object: ComputedSequenceOutputInfo,
    feature_dimensionalities_and_types: Dict[str, "FeatureExtractorInfo"],
    device: str,
) -> SequenceOutputModule:
    output_model_config = output_object.output_config.model_config
    output_module_type = output_model_config.model_type

    class_map = _get_sequence_output_module_type_class_map()
    cur_output_module_class = class_map[output_module_type]

    output_module = cur_output_module_class(
        output_object=output_object,
        output_name=output_object.output_config.output_info.output_name,
        feature_dimensionalities_and_types=feature_dimensionalities_and_types,
    )

    torch_device = torch.device(device=device)
    output_module = output_module.to(device=torch_device)

    return output_module


def _get_sequence_output_module_type_class_map() -> (
    dict[str, al_sequence_module_classes]
):
    mapping = {
        "sequence": SequenceOutputModule,
    }

    return mapping



from dataclasses import dataclass
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    Iterable,
    Protocol,
    Sequence,
    Tuple,
)

import numpy as np
import torch
from timm.data.mixup import rand_bbox

from eir.data_load.data_utils import Batch, get_output_info_generator
from eir.target_setup.target_label_setup import MissingTargetsInfo
from eir.train_utils.metrics import (
    FilteredOutputsAndLabels,
    filter_missing_outputs_and_labels,
)
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.setup.schemas import InputConfig
    from eir.train import Experiment, al_criteria_dict
    from eir.train_utils.criteria import al_losses
    from eir.train_utils.step_logic import al_training_labels_target

al_target_values = torch.LongTensor | torch.Tensor


logger = get_logger(name=__name__)


@dataclass
class MixingObject:
    ids: Sequence[str]
    targets: "al_training_labels_target"
    targets_permuted: "al_training_labels_target"
    lambda_: float
    permuted_indexes: torch.Tensor


def get_mix_data_hook(input_configs: Iterable["InputConfig"]):
    mixing_func_mapping = _get_mixing_func_map()

    input_mixing_func_map = {}

    for config in input_configs:
        cur_input_type_config = config.input_type_info
        cur_mixing_type = getattr(cur_input_type_config, "mixing_subtype", "mixup")

        cur_input_type = config.input_info.input_type
        cur_mixing_callable = mixing_func_mapping[cur_input_type][cur_mixing_type]

        cur_input_name = config.input_info.input_name

        input_mixing_func_map[cur_input_name] = cur_mixing_callable

    bound_hook = partial(
        hook_default_mix_data,
        mixing_funcs=input_mixing_func_map,
    )

    return bound_hook


class MixingCallable(Protocol):
    def __call__(
        self,
        tensor: torch.Tensor,
        lambda_: float,
        random_batch_indices_to_mix: torch.Tensor,
    ) -> torch.Tensor: ...


def _get_mixing_func_map() -> Dict[str, Dict[str, MixingCallable]]:
    mapping = {
        "omics": {
            "cutmix-uniform": uniform_cutmix_omics_input,
            "cutmix-block": block_cutmix_omics_input,
            "mixup": mixup_tensor,
        },
        "tabular": {"mixup": mixup_tensor},
        "sequence": {"mixup": mixup_tensor},
        "bytes": {"mixup": mixup_tensor},
        "image": {"mixup": mixup_tensor, "cutmix": cutmix_image},
        "array": {"mixup": mixup_tensor},
    }

    return mapping


def cutmix_image(
    tensor: torch.Tensor,
    lambda_: float,
    random_batch_indices_to_mix: torch.Tensor,
) -> torch.Tensor:
    image_shape = tensor[0].shape
    y_bottom, y_top, x_bottom, x_top = rand_bbox(img_shape=image_shape, lam=lambda_)

    target_to_cut = tensor[random_batch_indices_to_mix, :]
    cut_part = target_to_cut[..., y_bottom:y_top, x_bottom:x_top]

    # Caution: input_ will be modified as well since no .clone() below here
    cutmixed_x = tensor
    cutmixed_x[..., y_bottom:y_top, x_bottom:x_top] = cut_part

    return cutmixed_x


def hook_default_mix_data(
    experiment: "Experiment",
    state: Dict,
    mixing_funcs: Dict[str, MixingCallable],
    *args,
    **kwargs,
) -> dict[str, Any]:
    gc = experiment.configs.global_config

    batch = state["batch"]

    target_columns_gen = get_output_info_generator(outputs_as_dict=experiment.outputs)

    mixing_info = get_mixing_info(
        ids=batch.ids,
        mixing_alpha=gc.training_control.mixing_alpha,
        batch_size=gc.basic_experiment.batch_size,
        target_labels=batch.target_labels,
        target_columns_gen=target_columns_gen,
    )

    mixed_inputs = {}

    for input_name, input_data in batch.inputs.items():
        cur_mixing_func = mixing_funcs[input_name]
        mixed_input_batch = cur_mixing_func(
            tensor=input_data,
            lambda_=mixing_info.lambda_,
            random_batch_indices_to_mix=mixing_info.permuted_indexes,
        )
        mixed_inputs[input_name] = mixed_input_batch

    batch_mixed = Batch(
        inputs=mixed_inputs,
        target_labels=batch.target_labels,
        ids=batch.ids,
    )

    state_updates = {
        "batch": batch_mixed,
        "mixing_info": mixing_info,
    }

    return state_updates


def get_mixing_info(
    ids: Sequence[str],
    mixing_alpha: float,
    batch_size: int,
    target_labels: "al_training_labels_target",
    target_columns_gen: Generator[Tuple[str, str, str], None, None],
) -> MixingObject:
    lambda_ = _sample_lambda(mixing_alpha=mixing_alpha)

    permuted_indexes = get_random_batch_indices_to_mix(batch_size=batch_size)
    targets_permuted = mixup_all_targets(
        targets=target_labels,
        permuted_indices_for_mixing=permuted_indexes,
        target_columns_gen=target_columns_gen,
    )

    mixing_info = MixingObject(
        ids=ids,
        targets=target_labels,
        targets_permuted=targets_permuted,
        lambda_=lambda_,
        permuted_indexes=permuted_indexes,
    )

    return mixing_info


def _sample_lambda(mixing_alpha: float) -> float:
    if mixing_alpha > 0:
        beta_object = torch.distributions.beta.Beta(mixing_alpha, mixing_alpha)
        lambda_ = beta_object.sample().item()
    else:
        lambda_ = 1.0

    return lambda_


def hook_mix_loss(
    experiment: "Experiment",
    state: Dict,
    batch: "Batch",
    *args,
    **kwargs,
) -> Dict:
    target_columns_gen = get_output_info_generator(outputs_as_dict=experiment.outputs)

    valid_dataset = experiment.valid_dataset

    mixed_losses, filtered_outputs = calc_all_mixed_losses(
        target_columns_gen=target_columns_gen,
        criteria=experiment.criteria,
        outputs=state["model_outputs"],
        mixed_object=state["mixing_info"],
        missing_ids_per_output=valid_dataset.missing_ids_per_output,
    )

    state_updates = {
        "per_target_train_losses": mixed_losses,
        "filtered_outputs": filtered_outputs,
    }

    return state_updates


def get_random_batch_indices_to_mix(batch_size: int) -> torch.Tensor:
    return torch.randperm(batch_size).to(dtype=torch.long)


def mixup_tensor(
    tensor: torch.Tensor,
    lambda_: float,
    random_batch_indices_to_mix: torch.Tensor,
) -> torch.Tensor:
    mixed_tensor = (
        lambda_ * tensor + (1.0 - lambda_) * tensor[random_batch_indices_to_mix, :]
    )
    return mixed_tensor


def block_cutmix_omics_input(
    tensor: torch.Tensor, lambda_: float, random_batch_indices_to_mix: torch.Tensor
) -> torch.Tensor:
    cut_start, cut_end = get_block_cutmix_indices(
        input_length=tensor.shape[-1], lambda_=lambda_
    )
    target_to_cut = tensor[random_batch_indices_to_mix, :]
    cut_part = target_to_cut[..., cut_start:cut_end]

    # Caution: input_ will be modified as well since no .clone() below here
    cutmixed_x = tensor
    cutmixed_x[..., cut_start:cut_end] = cut_part

    return cutmixed_x


def get_block_cutmix_indices(input_length: int, lambda_: float) -> Tuple[int, int]:
    mixin_coefficient = 1.0 - lambda_
    num_snps_to_mix = int(round(input_length * mixin_coefficient))
    random_index_start = np.random.choice(max(1, input_length - num_snps_to_mix))
    random_index_end = random_index_start + num_snps_to_mix
    return random_index_start, random_index_end


def uniform_cutmix_omics_input(
    tensor: torch.Tensor,
    lambda_: float,
    random_batch_indices_to_mix: torch.Tensor,
) -> torch.Tensor:
    target_to_mix = tensor[random_batch_indices_to_mix, :]

    random_snp_indices_to_mix = get_uniform_cutmix_indices(
        input_length=tensor.shape[-1], lambda_=lambda_
    )
    cut_part = target_to_mix[..., random_snp_indices_to_mix]

    # Caution: input_ will be modified as well since no .clone() below here
    cutmixed_x = tensor
    cutmixed_x[..., random_snp_indices_to_mix] = cut_part

    return cutmixed_x


def get_uniform_cutmix_indices(input_length: int, lambda_: float) -> torch.Tensor:
    mixin_coefficient = 1.0 - lambda_
    num_snps_to_mix = int(round(input_length * mixin_coefficient))
    random_to_mix = np.random.choice(input_length, num_snps_to_mix, replace=False)
    random_to_mix_tensor = torch.tensor(random_to_mix, dtype=torch.long)

    return random_to_mix_tensor


def mixup_all_targets(
    targets: "al_training_labels_target",
    permuted_indices_for_mixing: torch.Tensor,
    target_columns_gen: Generator[Tuple[str, str, str], None, None],
) -> "al_training_labels_target":
    targets_permuted: "al_training_labels_target" = {}

    for output_name, target_type, target_name in target_columns_gen:
        if output_name not in targets_permuted:
            targets_permuted[output_name] = {}

        cur_targets = targets[output_name][target_name]
        cur_targets_permuted = mixup_targets(
            targets=cur_targets,
            permuted_indices_for_mixing=permuted_indices_for_mixing,
        )
        targets_permuted[output_name][target_name] = cur_targets_permuted

    return targets_permuted


def mixup_targets(
    targets: al_target_values,
    permuted_indices_for_mixing: torch.Tensor,
) -> al_target_values:
    targets_permuted = targets.detach().clone()
    targets_permuted = targets_permuted[permuted_indices_for_mixing]

    return targets_permuted


def calc_all_mixed_losses(
    target_columns_gen: Generator[tuple[str, str, str], None, None],
    criteria: "al_criteria_dict",
    outputs: dict[str, dict[str, torch.Tensor]],
    mixed_object: MixingObject,
    missing_ids_per_output: MissingTargetsInfo,
) -> tuple[dict[str, dict[str, torch.Tensor]], FilteredOutputsAndLabels]:
    losses: dict[str, dict[str, torch.Tensor]] = {
        output_name: {} for output_name in outputs.keys()
    }

    model_outputs = outputs
    target_labels = mixed_object.targets

    filtered_outputs = filter_missing_outputs_and_labels(
        batch_ids=list(mixed_object.ids),
        model_outputs=model_outputs,
        target_labels=target_labels,
        missing_ids_info=missing_ids_per_output,
        with_labels=True,
    )

    target_labels_permuted = mixed_object.targets_permuted
    ids_permuted = [mixed_object.ids[i] for i in mixed_object.permuted_indexes]
    filtered_outputs_permuted = filter_missing_outputs_and_labels(
        batch_ids=ids_permuted,
        model_outputs=model_outputs,
        target_labels=target_labels_permuted,
        missing_ids_info=missing_ids_per_output,
        with_labels=True,
    )
    filtered_targets_permuted = filtered_outputs_permuted.target_labels

    for output_name, target_type, target_name in target_columns_gen:
        cur_loss = calc_mixed_loss(
            criterion=criteria[output_name][target_name],
            outputs=filtered_outputs.model_outputs[output_name][target_name],
            targets=filtered_outputs.target_labels[output_name][target_name],
            targets_permuted=filtered_targets_permuted[output_name][target_name],
            lambda_=mixed_object.lambda_,
        )
        losses[output_name][target_name] = cur_loss

    return losses, filtered_outputs


def calc_mixed_loss(
    criterion: "al_losses",
    outputs: torch.Tensor,
    targets: torch.Tensor,
    targets_permuted: torch.Tensor,
    lambda_: float,
) -> torch.Tensor:
    base_loss = lambda_ * criterion(input=outputs, target=targets)
    permuted_loss = (1.0 - lambda_) * criterion(input=outputs, target=targets_permuted)

    total_loss = base_loss + permuted_loss

    return total_loss


def make_random_omics_columns_missing(
    omics_array: torch.Tensor,
    na_augment_alpha: float = 1.0,
    na_augment_beta: float = 9.0,
) -> torch.Tensor:
    if na_augment_alpha <= 0 or na_augment_beta <= 0:
        raise ValueError("Alpha and Beta must be positive.")

    dist = torch.distributions.beta.Beta(na_augment_alpha, na_augment_beta)
    percentage_sampled = dist.sample().item()

    n_snps = omics_array.shape[2]
    n_to_drop = int(n_snps * percentage_sampled)
    random_to_drop = torch.randperm(n_snps)[:n_to_drop].to(dtype=torch.long)

    missing_list = [
        False,
        False,
        False,
        True,
    ]
    missing_arr = torch.tensor(missing_list, dtype=torch.bool).reshape(-1, 1)
    omics_array[:, :, random_to_drop] = missing_arr

    return omics_array


def shuffle_random_omics_columns(
    omics_array: torch.Tensor,
    shuffle_augment_alpha: float = 1.0,
    shuffle_augment_beta: float = 19.0,
) -> torch.Tensor:
    if shuffle_augment_alpha <= 0 or shuffle_augment_beta <= 0:
        raise ValueError("Alpha and Beta must be positive.")

    dist = torch.distributions.beta.Beta(shuffle_augment_alpha, shuffle_augment_beta)
    percentage_sampled = dist.sample().item()

    n_snps = omics_array.shape[2]
    n_to_shuffle = int(n_snps * percentage_sampled)
    random_to_shuffle = torch.randperm(n_snps)[:n_to_shuffle].to(dtype=torch.long)

    one_hot_random = torch.zeros(
        omics_array.shape[0],
        4,
        n_to_shuffle,
        dtype=torch.bool,
    )

    random_indices = torch.randint(
        0,
        4,
        (omics_array.shape[0], n_to_shuffle),
    )
    one_hot_random.scatter_(1, random_indices.unsqueeze(1), 1)

    omics_array[:, :, random_to_shuffle] = one_hot_random

    return omics_array
